import inspect
import re

from functools import partial
from operator import attrgetter, getitem
from random import Random

from .base import TohuBaseGenerator, SeedGenerator
from .primitive_generators import as_tohu_generator

__all__ = ['Apply', 'GetAttribute', 'Lookup', 'SelectOneDerived', 'SelectMultipleDerived', 'TohuDict', 'fstr', 'ifthen']


class FuncArgGens:
    """
    Helper class
    """

    def __init__(self, arg_gens, kwarg_gens):
        self.arg_gens = arg_gens
        self.kwarg_gens = kwarg_gens
        self.seed_generator = SeedGenerator()

    def __next__(self):
        next_args = [next(g) for g in self.arg_gens]
        next_kwargs = {name: next(g) for name, g in self.kwarg_gens.items()}
        return next_args, next_kwargs

    def spawn(self):
        arg_gens_spawned = [g.spawn() for g in self.arg_gens]
        kwarg_gens_spawned = {name: g.spawn() for name, g in self.kwarg_gens.items()}
        return FuncArgGens(arg_gens_spawned, kwarg_gens_spawned)

    def clone(self):
        arg_gens_spawned = [g.clone() for g in self.arg_gens]
        kwarg_gens_spawned = {name: g.clone() for name, g in self.kwarg_gens.items()}
        return FuncArgGens(arg_gens_spawned, kwarg_gens_spawned)

    @property
    def all_generators(self):
        """
        Convenience property to iterate over all generators in arg_gens and kwarg_gens.
        """
        for arg_gen in self.arg_gens:
            yield arg_gen

        for kwarg_gen in self.kwarg_gens.values():
            yield kwarg_gen

    def reset(self, seed):
        self.seed_generator.reset(seed)

        for g in self.all_generators:
            g.reset(next(self.seed_generator))

    def _set_random_state_from(self, other):
        # Sanity checks
        assert isinstance(other, FuncArgGens)
        assert len(self.arg_gens) == len(other.arg_gens)
        assert len(self.kwarg_gens) == len(other.kwarg_gens)

        self.seed_generator._set_random_state_from(other.seed_generator)

        # Transfer random state from individual arg_gens
        for arg_gen_self, arg_gen_other in zip(self.arg_gens, other.arg_gens):
            arg_gen_self._set_random_state_from(arg_gen_other)

        # Transfer random state from individual kwarg_gens
        for name in self.kwarg_gens.keys():
            kwarg_gen_self = self.kwarg_gens[name]
            kwarg_gen_other = other.kwarg_gens[name]
            kwarg_gen_self._set_random_state_from(kwarg_gen_other)


class Apply(TohuBaseGenerator):

    def __init__(self, func, *arg_gens, **kwarg_gens):
        super().__init__()
        self._check_func_is_callable(func)
        self.func = func
        self.func_arg_gens_orig = FuncArgGens(arg_gens, kwarg_gens)
        self.func_arg_gens_internal = self.func_arg_gens_orig.clone()

    def _check_func_is_callable(self, func):
        if not callable(func):
            raise TypeError("First argument must be callable. Got: {func} (type: {type(func)})")

    def __next__(self):
        next_args, next_kwargs = next(self.func_arg_gens_internal)
        return self.func(*next_args, **next_kwargs)

    def reset(self, seed):
        super().reset(seed)

    def spawn(self):
        new_obj = Apply(self.func, *self.func_arg_gens_orig.arg_gens, **self.func_arg_gens_orig.kwarg_gens)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        #self.func_arg_gens_orig._set_random_state_from(other.func_arg_gens_orig)
        self.func_arg_gens_internal._set_random_state_from(other.func_arg_gens_internal)


class GetAttribute(Apply):

    def __init__(self, parent, name):
        self.parent = parent  # no need to clone here because this happens in the superclass
        self.name = name
        func = attrgetter(self.name)
        super().__init__(func, self.parent)

    def spawn(self):
        return GetAttribute(self.parent, self.name)

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)


class Lookup(Apply):

    def __init__(self, parent, mapping):
        self.parent = parent  #  no need to clone here because this happens in the superclass
        self.mapping = mapping
        func = partial(getitem, self.mapping)
        super().__init__(func, parent)

    def spawn(self):
        return Lookup(self.parent, self.mapping)

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)


class SelectOneDerived(Apply):

    def __init__(self, parent, p=None):
        self.parent = as_tohu_generator(parent)
        self.p = as_tohu_generator(p)
        self.randgen = Random()
        func = self.randgen.choice
        super().__init__(func, parent)

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)

    def spawn(self):
        return SelectOneDerived(self.parent, self.p)

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.randgen.setstate(other.randgen.getstate())


# TODO: find a better name for this class!
class SelectMultipleDerived(Apply):

    def __init__(self, parent, num):
        assert isinstance(parent, TohuBaseGenerator)
        assert isinstance(num, TohuBaseGenerator)
        self.parent = parent
        self.num = num
        self.randgen = Random()
        func = self.randgen.choices
        super().__init__(func, parent, k=self.num)

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)

    def spawn(self):
        return SelectMultipleDerived(self.parent, self.num)

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.randgen.setstate(other.randgen.getstate())


class fstr(Apply):
    """
    Helper function for easy formatting of generators.

    Usage example:

        >>> g1 = Integer(100, 200)
        >>> g2 = Integer(300, 400)
        >>> g3 = g1 + g2
        >>> h = fstr('{g1} + {g2} = {g3}')
        >>> print(next(h))
        122 + 338 = 460
        >>> print(next(h))
        165 + 325 = 490
    """

    def __init__(self, spec, namespace=None):

        # FIXME: this pattern is not yet compatible with the full f-string spec.
        # For example, it doesn't recognise double '{{' and '}}' (for escaping).
        # Also it would be awesome if we could parse arbitrary expressions inside
        # the curly braces.
        pattern = '{([^}]+)}'

        gen_names = re.findall(pattern, spec)

        if namespace is None:
            namespace = inspect.currentframe().f_back.f_globals
            namespace.update(inspect.currentframe().f_back.f_locals)

        gens = {name: namespace[name] for name in gen_names}

        def format_items(**kwargs):
            return spec.format(**kwargs)

        super().__init__(format_items, **gens)

    def spawn(self):
        raise NotImplemented("TODO: check which namespace to use when spawning an instance of 'fstr'")


class TohuDict:
    """
    Helper class which behaves like a regular dictionary but
    also allows easy lookup of items produced by a generator.
    """

    def __init__(self, mapping):
        self.mapping = mapping

    def __getitem__(self, key):
        if isinstance(key, TohuBaseGenerator):
            return Lookup(key, self.mapping)
        else:
            return self.mapping[key]


class ifthen(Apply):

    def __init__(self, g_cond, g_true, g_false):
        """
        Parameters
        ----------
        g_cond: tohu.generator

        :param g_cond:
        :param g_true:
        :param g_false:
        """
        self.g_cond = g_cond
        self.g_true = g_true
        self.g_false = g_false

        def func(cond, x, y):
            return x if cond else y

        super().__init__(func, self.g_cond, as_tohu_generator(self.g_true), as_tohu_generator(self.g_false))

    def spawn(self):
        return ifthen(self.g_cond, self.g_true, self.g_false)



from abc import ABCMeta
from ..base import TohuBaseGenerator
from ..tohu_namespace import TohuNamespace
from .utils import make_tohu_items_class, get_tohu_items_name

__all__ = ['CustomGenerator']


def augment_init_method(cls):
    """
    Replace the existing cls.__init__() method with a new one which
    also initialises the field generators and similar bookkeeping.
    """

    orig_init = cls.__init__

    def new_init(self, *args, **kwargs):
        super(CustomGenerator, self).__init__()  # TODO: does this behave correctly with longer inheritance chains?

        orig_init(self, *args, **kwargs)

        self.orig_args = args
        self.orig_kwargs = kwargs

        self.ns_gen_templates = TohuNamespace()
        self.ns_gen_templates.update_from_dict(self.__class__.__dict__)
        self.ns_gen_templates.update_from_dict(self.__dict__)
        self.ns_gen_templates.set_owner(self.__class__)
        self._mark_field_generator_templates()

        self.ns_gens = self.ns_gen_templates.spawn()
        self.ns_gens.set_owner(self)

        self._update_namespace_with_field_generators()
        self._set_field_names()
        self._set_tohu_items_name()
        self._set_tohu_items_cls()

    cls.__init__ = new_init


class CustomGeneratorMeta(ABCMeta):

    def __new__(metacls, cg_name, bases, clsdict):
        # Create new custom generator class
        new_cls = super(CustomGeneratorMeta, metacls).__new__(metacls, cg_name, bases, clsdict)

        # Augment original init method with bookkeeping needed for custom generators
        augment_init_method(new_cls)

        return new_cls


class CustomGenerator(TohuBaseGenerator, metaclass=CustomGeneratorMeta):

    def _mark_field_generator_templates(self):
        """
        Mark field generator templates as such so that an indication of this is
        included in the tohu_name. This is purely convenience for easier debugging.
        """
        for g in self.ns_gen_templates.all_generators:
            g.is_custom_generator_template = True

    def _update_namespace_with_field_generators(self):
        self.__dict__.update(self.ns_gens.named_generators)

    def _set_field_names(self):
        constituent_generator_names = list(self.ns_gens.names)

        if not hasattr(self, '__fields__'):
            self.field_names = constituent_generator_names
        else:
            self.field_names = self.__fields__

            # sanity check
            for field_name in self.field_names:
                if field_name not in constituent_generator_names:
                    raise ValueError(f"Attribute __fields__ contains name which is not a named field generator: '{field_name}'")

    def _set_tohu_items_name(self):
        self.__class__.__tohu_items_name__ = get_tohu_items_name(self.__class__)

    def _set_tohu_items_cls(self):
        if not hasattr(self.__class__, 'tohu_items_cls'):
            self.__class__.tohu_items_cls = make_tohu_items_class(self.__tohu_items_name__, self.field_names)

    def __next__(self):
        field_values = {name: next(self.ns_gens[name]) for name in self.field_names}
        return self.tohu_items_cls(**field_values)

    def reset(self, seed):
        super().reset(seed)
        self.ns_gens.reset(seed)
        return self

    def spawn(self, spawn_mapping=None):
        new_obj = self.__class__(*self.orig_args, **self.orig_kwargs)
        new_obj._set_random_state_from(self)

        # # Explicitly set tohu_items_cls. This is necessary because due to
        # # the way in which `attr` works, explicit comparisons between
        # # generated items will return False even though they contain
        # # the same elements (because the underlying attr classes are
        # # different, so attr plays it safe).
        # new_obj.tohu_items_cls = self.tohu_items_cls

        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.ns_gens._set_random_state_from(other.ns_gens)



import attr
import datetime as dt
import geojson
import numpy as np
import shapely

from faker import Faker
from functools import partial
from random import Random
from shapely.geometry import Point, Polygon, MultiPolygon

from .base import TohuBaseGenerator, SeedGenerator
from .item_list import ItemList
from .logging import logger
from .utils import identity

__all__ = ['Boolean', 'CharString', 'Constant', 'DigitString', 'FakerGenerator', 'Float', 'GeoJSONGeolocation',
           'HashDigest', 'Integer', 'IterateOver', 'NumpyRandomGenerator', 'SelectOnePrimitive',
           'SelectMultiplePrimitive', 'Sequential', 'Timestamp', 'as_tohu_generator']


class PrimitiveGenerator(TohuBaseGenerator):
    """
    Base class for all primitive generators
    """


class Constant(PrimitiveGenerator):
    """
    Generator which produces a constant sequence (repeating the same value indefinitely).
    """

    def __init__(self, value):
        """
        Parameters
        ----------
        value:
            The constant value produced by this generator.
        """
        super().__init__()
        self.value = value

    def reset(self, seed=None):
        super().reset(seed)
        return self

    def __next__(self):
        return self.value

    def spawn(self):
        return Constant(self.value)

    def _set_random_state_from(self, other):
        pass


class Boolean(PrimitiveGenerator):
    """
    Generator which produces random boolean values (True or False).
    """

    def __init__(self, p=0.5):
        """
        Parameters
        ----------
        p: float
            The probability that True is returned. Must be between 0.0 and 1.0.
        """
        super().__init__()
        self.p = p
        self.randgen = Random()

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)
        return self

    def __next__(self):
        return self.randgen.random() < self.p

    def spawn(self):
        new_obj = Boolean(self.p)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.randgen.setstate(other.randgen.getstate())


class Integer(PrimitiveGenerator):
    """
    Generator which produces random integers k in the range low <= k <= high.
    """

    def __init__(self, low, high):
        """
        Parameters
        ----------
        low: integer
            Lower bound (inclusive).
        high: integer
            Upper bound (inclusive).
        """
        super().__init__()
        self.low = low
        self.high = high
        self.randgen = Random()

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)
        return self

    def __next__(self):
        return self.randgen.randint(self.low, self.high)

    def spawn(self):
        new_obj = Integer(self.low, self.high)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.randgen.setstate(other.randgen.getstate())


class Float(PrimitiveGenerator):
    """
    Generator which produces random floating point numbers x in the range low <= x <= high.
    """

    def __init__(self, low, high):
        """
        Parameters
        ----------
        low: integer
            Lower bound (inclusive).
        high: integer
            Upper bound (inclusive).
        """
        super().__init__()
        self.low = low
        self.high = high
        self.randgen = Random()

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)
        return self

    def __next__(self):
        return self.randgen.uniform(self.low, self.high)

    def spawn(self):
        new_obj = Float(self.low, self.high)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.randgen.setstate(other.randgen.getstate())


CHARACTER_SETS = {
    '<alphanumeric>': 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789',
    '<alphanumeric_lowercase>': 'abcdefghijklmnopqrstuvwxyz0123456789',
    '<alphanumeric_uppercase>': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789',
    '<lowercase>': 'abcdefghijklmnopqrstuvwxyz',
    '<uppercase>': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
    '<digits>': '0123456789',
}


class CharString(PrimitiveGenerator):
    """
    Generator which produces a sequence of character strings.
    """

    def __init__(self, *, length, charset='<alphanumeric>'):
        """
        Parameters
        ----------
        length: integer
            Length of the character strings produced by this generator.
        charset: iterable
            Character set to draw from when generating strings, or string
            with the name of a pre-defined character set.
            Default: <alphanumeric> (both lowercase and uppercase letters).
        """
        super().__init__()
        self.length = length
        try:
            self.charset = CHARACTER_SETS[charset]
            logger.debug(f"Using pre-defined character set: '{charset}'")
        except KeyError:
            self.charset = charset
        self.seed_generator = SeedGenerator()
        self.char_gen = Random()

    def spawn(self):
        new_obj = CharString(length=self.length, charset=self.charset)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.seed_generator._set_random_state_from(other.seed_generator)
        self.char_gen.setstate(other.char_gen.getstate())

    def __next__(self):
        chars = self.char_gen.choices(self.charset, k=self.length)
        return ''.join(chars)

    def reset(self, seed):
        super().reset(seed)
        self.seed_generator.reset(seed)
        self.char_gen.seed(next(self.seed_generator))
        return self


class DigitString(CharString):
    """
    Generator which produces a sequence of strings containing only digits.
    """

    def __init__(self, *, length=None):
        """
        Parameters
        ----------
        length: integer
            Length of the character strings produced by this generator.
        """
        charset = "0123456789"
        super().__init__(length=length, charset=charset)

    def spawn(self):
        new_obj = DigitString(length=self.length)
        new_obj._set_random_state_from(self)
        return new_obj


class HashDigest(PrimitiveGenerator):
    """
    Generator which produces a sequence of hex strings representing hash digest values.
    """

    def __init__(self, *, length=None, as_bytes=False, uppercase=True):
        """
        Parameters
        ----------
        length: integer
            Length of the character strings produced by this generator.
        as_bytes: bool
            If True, return `length` random bytes. If False, return a string of `length`
            characters with a hexadecimal representation of `length/2` random bytes.
            Note that in the second case `length` must be an even number.
        uppercase: bool
            If True (the default), return hex string using uppercase letters, otherwise lowercase.
            This only has an effect if `as_bytes=False`.
        """
        super().__init__()
        self.length = length
        self._internal_length = length if as_bytes else length / 2
        if not as_bytes and (length % 2) != 0:
            raise ValueError(
                f"Length must be an even number if as_bytes=False because it "
                f"represents length = 2 * num_random_bytes. Got: length={length})")
        self.as_bytes = as_bytes
        self.uppercase = uppercase
        self.randgen = np.random.RandomState()
        self._maybe_convert_to_hex = identity if self.as_bytes else bytes.hex
        self._maybe_convert_to_uppercase = identity if (self.as_bytes or not uppercase) else str.upper

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)
        return self

    def __next__(self):
        val = self.randgen.bytes(self._internal_length)
        return self._maybe_convert_to_uppercase(self._maybe_convert_to_hex(val))

    def spawn(self):
        new_obj = HashDigest(length=self.length, as_bytes=self.as_bytes, uppercase=self.uppercase)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.randgen.set_state(other.randgen.get_state())


class Sequential(PrimitiveGenerator):
    """
    Generator which produces a sequence of strings
    of the form:

        "PREFIX001"
        "PREFIX002"
        "PREFIX003"
        ...

    Both the prefix and the number of digits can
    be modified by the user.

    Example:
        >>> s = Sequential(prefix="Foobar_", digits=4)
        >>> next(s)
        Foobar_0001
        >>> next(s)
        Foobar_0002
        >>> next(s)
        Foobar_0003
    """

    def __init__(self, *, prefix, digits):
        """
        Parameters
        ----------
        prefix: string
            Prefix to be appended to generated elements.
        digits: integer
            Number of digits to use for the sequential numbering.
            Any numbers will fewer digits will be zero-padded;
            numbers with more digits are unaffected.
        """
        super().__init__()
        self.prefix = prefix
        self.digits = digits
        self.fmt_str = self.prefix + '{{:0{digits}}}'.format(digits=digits)
        self.cnt = 0

    def spawn(self):
        new_obj = Sequential(prefix=self.prefix, digits=self.digits)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.cnt = other.cnt

    def reset(self, seed=None):
        """
        Note that this method supports the `seed` argument (for consistency with other generators),
        but its value is ignored - the generator is simply reset to its initial value.
        """
        super().reset(seed)
        self.cnt = 0
        return self

    def __next__(self):
        self.cnt += 1
        return self.fmt_str.format(self.cnt)


class NumpyRandomGenerator(TohuBaseGenerator):
    """
    Generator which produces random numbers using one of the methods supported by numpy. [1]

    [1] https://docs.scipy.org/doc/numpy/reference/routines.random.html
    """

    def __init__(self, method, **numpy_args):
        """
        Parameters
        ----------
        method: string
            Name of the numpy function to use (see [1] for details)
        numpy_args:
            Remaining arguments passed to the numpy function (see [1] for details)

        References
        ----------
        [1] https://docs.scipy.org/doc/numpy/reference/routines.random.html
        """
        super().__init__()
        self.method = method
        self.random_state = np.random.RandomState()
        self.randgen = getattr(self.random_state, method)
        self.numpy_args = numpy_args

    def reset(self, seed):
        super().reset(seed)
        self.random_state.seed(seed)
        return self

    def __next__(self):
        return self.randgen(**self.numpy_args)

    def spawn(self):
        new_obj = NumpyRandomGenerator(method=self.method, **self.numpy_args)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.random_state.set_state(other.random_state.get_state())


class FakerGenerator(PrimitiveGenerator):
    """
    Generator which produces random elements using one of the methods supported by faker. [1]

    [1] https://faker.readthedocs.io/
    """

    def __init__(self, method, *, locale=None, **faker_args):
        """
        Parameters
        ----------
        method: string
            Name of the faker provider to use (see [1] for details)
        locale: string
             Locale to use when generating data, e.g. 'en_US' (see [1] for details)
        faker_args:
            Remaining arguments passed to the faker provider (see [1] for details)

        References
        ----------
        [1] https://faker.readthedocs.io/
        """
        super().__init__()
        self.method = method
        self.locale = locale
        self.faker_args = faker_args

        self.fake = Faker(locale=locale)
        self.randgen = getattr(self.fake, method)
        self.fake.seed_instance(None)  # seed instance to ensure we are decoupled from the global random state

    def reset(self, seed):
        super().reset(seed)
        self.fake.seed_instance(seed)
        return self

    def __next__(self):
        return self.randgen(**self.faker_args)

    def spawn(self):
        new_obj = FakerGenerator(self.method, locale=self.locale, **self.faker_args)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.fake.random.setstate(other.fake.random.getstate())



class IterateOver(PrimitiveGenerator):
    """
    Generator which simply iterates over all items in a given iterable
    """

    def __init__(self, seq):
        if not isinstance(seq, (list, tuple, ItemList, str)):
            raise TypeError(
                f"For the time being 'seq' must be a list, tuple, ItemList or string "
                f"so that we can reproducibly spawn and reset this generator. Got: {seq}")

        super().__init__()
        self.seq = seq
        # Note: iterating using an explicit index isn't ideal but it allows
        # to transfer the internal state when spawning (for reproducibility)
        self.idx = 0
        self.reset()

    def __repr__(self):
        return f"<IterateOver, list with {len(self.seq)} items>"

    def __next__(self):
        try:
            val = self.seq[self.idx]
        except IndexError:
            raise StopIteration()
        self.idx += 1
        return val

    def __iter__(self):
        return self

    def reset(self, seed=None):
        super().reset(seed)
        self.idx = 0
        return self

    def spawn(self):
        new_obj = IterateOver(self.seq)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.idx = other.idx


class SelectOnePrimitive(PrimitiveGenerator):
    """
    Generator which produces a sequence of items taken from a given set of elements.
    """

    def __init__(self, values, p=None):
        """
        Parameters
        ----------
        values: list
            List of options from which to choose elements.
        p: list, optional
            The probabilities associated with each element in `values`.
            If not given the assumes a uniform distribution over all values.
        """
        super().__init__()
        self.values = list(values)  # need to convert to a list so that numpy.random.RandomState() doesn't get confused
        self.p = p
        self.randgen = None
        self.func_random_choice = None
        self._init_randgen()

    def _init_randgen(self):
        """
        Initialise random generator to be used for picking elements.
        With the current implementation in tohu (where we pick elements
        from generators individually instead of in bulk), it is faster
        to `use random.Random` than `numpy.random.RandomState` (it is
        possible that this may change in the future if we change the
        design so that tohu pre-produces elements in bulk, but that's
        not likely to happen in the near future).

        Since `random.Random` doesn't support arbitrary distributions,
        we can only use it if `p=None`. This helper function returns
        the appropriate random number generator depending in the value
        of `p`, and also returns a function `random_choice` which can be
        applied to the input sequence to select random elements from it.
        """
        if self.p is None:
            self.randgen = Random()
            self.func_random_choice = self.randgen.choice
        else:
            self.randgen = np.random.RandomState()
            self.func_random_choice = partial(self.randgen.choice, p=self.p)

    def _set_random_state_from(self, other):
        """
        Transfer the internal state from `other` to `self`.
        After this call, `self` will produce the same elements
        in the same order as `other` (even though they otherwise
        remain completely independent).
        """
        try:
            # this works if randgen is an instance of random.Random()
            self.randgen.setstate(other.randgen.getstate())
        except AttributeError:
            # this works if randgen is an instance of numpy.random.RandomState()
            self.randgen.set_state(other.randgen.get_state())

        return self

    def __next__(self):
        return self.func_random_choice(self.values)

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)
        return self

    def spawn(self):
        new_obj = SelectOnePrimitive(self.values, p=self.p)
        new_obj._set_random_state_from(self)
        return new_obj


class SelectMultiplePrimitive(PrimitiveGenerator):
    """
    Generator which produces a sequence of items taken from a given set of elements.
    """

    def __init__(self, values, num, p=None):
        """
        Parameters
        ----------
        values: list
            List of options from which to choose elements.
        num: int
            Number of elements to select.
        p: list, optional
            The probabilities associated with each element in `values`.
            If not given the assumes a uniform distribution over all values.
        """
        super().__init__()
        self.values = list(values)  # need to convert to a list so that numpy.random.RandomState() doesn't get confused
        self.num = num
        self.p = p
        self.randgen = None
        self.func_random_choice = None
        self._init_randgen()

    def _init_randgen(self):
        """
        Initialise random generator to be used for picking elements.
        With the current implementation in tohu (where we pick elements
        from generators individually instead of in bulk), it is faster
        to `use random.Random` than `numpy.random.RandomState` (it is
        possible that this may change in the future if we change the
        design so that tohu pre-produces elements in bulk, but that's
        not likely to happen in the near future).

        Since `random.Random` doesn't support arbitrary distributions,
        we can only use it if `p=None`. This helper function returns
        the appropriate random number generator depending in the value
        of `p`, and also returns a function `random_choice` which can be
        applied to the input sequence to select random elements from it.
        """
        if self.p is None:
            self.randgen = Random()
            self.func_random_choice = partial(self.randgen.choices, k=self.num)
        else:
            self.randgen = np.random.RandomState()
            self.func_random_choice = partial(self.randgen.choice, p=self.p, k=self.num)

    def _set_random_state_from(self, other):
        """
        Transfer the internal state from `other` to `self`.
        After this call, `self` will produce the same elements
        in the same order as `other` (even though they otherwise
        remain completely independent).
        """
        try:
            # this works if randgen is an instance of random.Random()
            self.randgen.setstate(other.randgen.getstate())
        except AttributeError:
            # this works if randgen is an instance of numpy.random.RandomState()
            self.randgen.set_state(other.randgen.get_state())

        return self

    def __next__(self):
        return self.func_random_choice(self.values)

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)
        return self

    def spawn(self):
        new_obj = SelectMultiplePrimitive(self.values, num=self.num, p=self.p)
        new_obj._set_random_state_from(self)
        return new_obj


def as_tohu_generator(g):
    """
    If g is a tohu generator return it unchanged,
    otherwise wrap it in a Constant generator.
    """

    if isinstance(g, TohuBaseGenerator):
        return g
    else:
        return Constant(g)


class ShapelyGeolocation(PrimitiveGenerator):
    """
    Generator which produces random locations inside a shapely polygon
    or multipolygon. This is a helper class and most users will probably
    find the GeoJSONGeolocation generator more useful.
    """

    def __init__(self, shp, properties=None, max_tries=100):
        if not isinstance(shp, (Polygon, MultiPolygon)):
            raise TypeError(f"Argument 'shp' must be of type Polygon or MultiPolygon. Got: {type(shp)}")

        super().__init__()

        self.shape = shapely.geometry.shape(shp)
        self.properties = properties or dict()

        self.geolocation_cls = self._make_geolocation_class()

        lon_min, lat_min, lon_max, lat_max = self.shape.bounds
        self.lon_gen = Float(lon_min, lon_max)
        self.lat_gen = Float(lat_min, lat_max)
        self.max_tries = max_tries
        self.seed_generator = SeedGenerator()

    def _make_geolocation_class(self):
        fields = {'lon': attr.ib(), 'lat': attr.ib()}
        fields.update({name: attr.ib(value) for name, value in self.properties.items()})
        cls = attr.make_class('Geolocation', fields)
        cls.as_dict = lambda self: attr.asdict(self)
        def __new_eq__(self, other):
            return self.lon == other.lon and self.lat == other.lat
        cls.__eq__ = __new_eq__
        return cls

    def __repr__(self):
        return f"<ShapelyShape, area={self.area:.3f}>"

    def spawn(self):
        new_obj = ShapelyGeolocation(self.shape, properties=self.properties, max_tries=self.max_tries)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.seed_generator._set_random_state_from(other.seed_generator)
        self.lon_gen._set_random_state_from(other.lon_gen)
        self.lat_gen._set_random_state_from(other.lat_gen)

    @property
    def area(self):
        return self.shape.area

    def __next__(self):
        for cnt in range(1, self.max_tries + 1):
            pt = Point(next(self.lon_gen), next(self.lat_gen))
            if pt.within(self.shape):
                return self.geolocation_cls(lon=pt.x, lat=pt.y)
            else:
                logger.debug(f"Generated point is not within shape. Trying again... [{cnt}/{self.max_tries}]")
        raise RuntimeError(f"Could not generate point in shape after {self.max_tries} attempts")

    def reset(self, seed):
        super().reset(seed)
        self.seed_generator.reset(seed)
        self.lon_gen.reset(next(self.seed_generator))
        self.lat_gen.reset(next(self.seed_generator))
        return self


class GeoJSONGeolocation(PrimitiveGenerator):
    """
    Generator which produces random locations inside a geographic area.
    """

    def __init__(self, filename_or_geojson_data, include_attributes=None, max_tries=100):
        super().__init__()

        if isinstance(filename_or_geojson_data, str):
            try:
                with open(filename_or_geojson_data, 'r') as f:
                    geojson_data = geojson.load(f)
            except AttributeError:
                raise NotImplementedError()
        else:
            geojson_data = filename_or_geojson_data

        self.geojson_data = geojson_data
        self.include_attributes = include_attributes or []
        self.max_tries = max_tries

        self.shape_gens = self._make_shape_generators()

        areas = np.array([s.area for s in self.shape_gens])
        self.choice_probs = areas / areas.sum()  # TODO: allow weighin by an arbitrary attribute, not just by area

        self.seed_generator = SeedGenerator()
        self.shape_gen_chooser = np.random.RandomState()

    def _make_shape_generators(self):
        shape_gens = []

        for feature in self.geojson_data['features']:

            geom = shapely.geometry.shape(feature['geometry'])

            cur_attributes = {}
            for name in self.include_attributes:
                try:
                    cur_attributes[name] = feature['properties'][name]
                except KeyError:
                    valid_attributes = list(feature['properties'].keys())
                    raise ValueError(f"Feature does not have attribute '{name}'. Valid attributes are: {valid_attributes}")

            shape_gens.append(ShapelyGeolocation(geom, cur_attributes, max_tries=self.max_tries))

        return shape_gens

    def spawn(self):
        new_obj = GeoJSONGeolocation(self.geojson_data, include_attributes=self.include_attributes, max_tries=self.max_tries)
        new_obj._set_random_state_from(self)
        return new_obj

    def __next__(self):
        sg = self.shape_gen_chooser.choice(self.shape_gens, p=self.choice_probs)
        return next(sg)

    def reset(self, seed):
        super().reset(seed)
        self.seed_generator.reset(seed)
        self.shape_gen_chooser.seed(next(self.seed_generator))
        for g in self.shape_gens:
            g.reset(next(self.seed_generator))

    def _set_random_state_from(self, other):
        self.seed_generator._set_random_state_from(other.seed_generator)
        self.shape_gen_chooser.set_state(other.shape_gen_chooser.get_state())
        for gen_self, gen_other in zip(self.shape_gens, other.shape_gens):
            gen_self._set_random_state_from(gen_other)


class TimestampError(Exception):
    """
    Custom exception raised to indicate Timestamp errors
    (for example when the end time of the timestamp interval
    is before the start time).
    """


class Timestamp(PrimitiveGenerator):
    """
    Generator which produces random timestamps.
    """

    def __init__(self, *, start=None, end=None, date=None, fmt=None, uppercase=False):
        """
        Initialise timestamp generator.

        Note that `start` and `end` are both inclusive. They can either
        be full timestamps such as 'YYYY-MM-DD HH:MM:SS', or date strings
        such as 'YYYY-MM-DD'. Note that in the latter case `end` is
        interpreted as as `YYYY-MM-DD 23:59:59`, i.e. the day is counted
        in full.

        The produced timestamps are datetime objects, but calling
        `strftime(fmt=...)` on this generator another generator is
        returned which produces timestamps as strings instead.

        Alternatively, you can provide the `fmt` argument directly
        to produce timestamps formatted as strings.

        Args:
            start (date string):  start time
            end   (date string):  end time
            date (str):           string of the form YYYY-MM-DD. This is an alternative
                                  (and mutually exclusive) to specifying `start` and `end`.
                                  It is equivalent to setting start='YYYY-MM-DD 00:00:00',
                                  end='YYYY-MM-DD 23:59:59'.
            fmt (str or None):    if given, return timestamps as strings instead of datetime
                                  objects, using `fmt` as the formatting string (default: None)
            uppercase (bool):     whether to format strings as uppercase. If `fmt` is None
                                  then this parameter is ignored.
        """
        super().__init__()

        if (date is not None):
            if not (start is None and end is None):
                raise TimestampError("Argument `date` is mutually exclusive with `start` and `end`.")

            # ensure that date is a string of the form 'YYYY-MM-DD'
            try:
                date = date.strftime('%Y-%m-%d')
            except AttributeError:
                if not isinstance(date, str):
                    raise TypeError("Argument `date` must be either a string or a date-like object "
                                    f"(e.g. datetime.date or pandas.Timestamp). Got: {type(date)}")

            self.start = dt.datetime.strptime(date, '%Y-%m-%d')
            self.end = self.start + dt.timedelta(hours=23, minutes=59, seconds=59)
        else:
            if (start is None or end is None):
                raise TimestampError("Either `date` or both `start` and `end` must be provided.")

            try:
                self.start = dt.datetime.strptime(start, '%Y-%m-%d')
            except ValueError:
                self.start = dt.datetime.strptime(start, '%Y-%m-%d %H:%M:%S')

            try:
                self.end = dt.datetime.strptime(end, '%Y-%m-%d %H:%M:%S')
            except ValueError:
                end_date = dt.datetime.strptime(end, '%Y-%m-%d')
                self.end = dt.datetime(end_date.year, end_date.month, end_date.day, 23, 59, 59)

        self.dt = int((self.end - self.start).total_seconds())

        if self.dt < 0:
            raise TimestampError(f"Start time must be before end time. Got: start_time='{self.start}', end_time='{self.end}'.")

        self.fmt = fmt
        self.uppercase = uppercase

        if self.fmt is None:
            self._maybe_format_timestamp = identity
        else:
            if not isinstance(self.fmt, str):
                raise ValueError(f"Argument 'fmt' must be of type string, got '{type(self.fmt)}'")

            if uppercase:
                self._maybe_format_timestamp = lambda ts: ts.strftime(self.fmt).upper()
            else:
                self._maybe_format_timestamp = lambda ts: ts.strftime(self.fmt)

        self.offset_randgen = Random()

    def spawn(self):
        new_obj = Timestamp(start=self.start.strftime('%Y-%m-%d %H:%M:%S'), end=self.end.strftime('%Y-%m-%d %H:%M:%S'),
                            fmt=self.fmt, uppercase=self.uppercase)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.offset_randgen.setstate(other.offset_randgen.getstate())

    def __next__(self):
        next_offset = self.offset_randgen.randint(0, self.dt)
        ts = (self.start + dt.timedelta(seconds=next_offset))
        return self._maybe_format_timestamp(ts)

    def reset(self, seed):
        super().reset(seed)
        self.offset_randgen.seed(seed)
        return self

    def strftime(self, fmt='%Y-%m-%d %H:%M:%S', uppercase=False):
        g = Timestamp(start=self.start.strftime('%Y-%m-%d %H:%M:%S'), end=self.end.strftime('%Y-%m-%d %H:%M:%S'),
                      fmt=fmt, uppercase=uppercase)
        self.register_clone(g)
        return g


import attr
import pandas as pd
import re

from ..base import TohuBaseGenerator
from ..logging import logger

__all__ = ['get_tohu_items_name', 'make_tohu_items_class']


def make_tohu_items_class(clsname, attr_names):
    """
    Parameters
    ----------
    clsname: string
        Name of the class to be created

    attr_names: list of strings
        Names of the attributes of the class to be created
    """

    item_cls = attr.make_class(clsname, {name: attr.ib() for name in attr_names}, repr=False, cmp=True, frozen=True)

    def new_repr(self):
        all_fields = ', '.join([f'{name}={repr(value)}' for name, value in attr.asdict(self).items()])
        return f'{clsname}({all_fields})'

    orig_eq = item_cls.__eq__

    def new_eq(self, other):
        """
        Custom __eq__() method which also allows comparisons with
        tuples and dictionaries. This is mostly for convenience
        during testing.
        """

        if isinstance(other, self.__class__):
            return orig_eq(self, other)
        else:
            if isinstance(other, tuple):
                return attr.astuple(self) == other
            elif isinstance(other, dict):
                return attr.asdict(self) == other
            else:
                return NotImplemented

    item_cls.__repr__ = new_repr
    item_cls.__eq__ = new_eq
    item_cls.keys = lambda self: attr_names
    item_cls.__getitem__ = lambda self, key: getattr(self, key)
    item_cls.as_dict = lambda self: attr.asdict(self)
    item_cls.to_series = lambda self: pd.Series(attr.asdict(self))

    return item_cls


def get_tohu_items_name(cls):
    """
    Return a string which defines the name of the namedtuple class which will be used
    to produce items for the custom generator.

    By default this will be the first part of the class name (before '...Generator'),
    for example:

        FoobarGenerator -> Foobar
        QuuxGenerator   -> Quux

    However, it can be set explicitly by the user by defining `__tohu_items_name__`
    in the class definition, for example:

        class Quux(CustomGenerator):
            __tohu_items_name__ = 'MyQuuxItem'
    """
    assert issubclass(cls, TohuBaseGenerator)

    try:
        tohu_items_name = cls.__dict__['__tohu_items_name__']
        logger.debug(f"Using item class name '{tohu_items_name}' (derived from attribute '__tohu_items_name__')")
    except KeyError:
        m = re.match('^(.*)Generator$', cls.__name__)
        if m is not None:
            tohu_items_name = m.group(1)
            logger.debug(f"Using item class name '{tohu_items_name}' (derived from custom generator name)")
        else:
            msg = (
                "Cannot derive class name for items to be produced by custom generator. "
                "Please set '__tohu_items_name__' at the top of the custom generator's "
                "definition or change its name so that it ends in '...Generator'"
            )
            raise ValueError(msg)

    return tohu_items_name


import logging

__all__ = ['logger']

#
# Create logger
#
logger = logging.getLogger('tohu_v6')
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('{asctime} {levelname}  {message}', datefmt='%Y-%m-%d %H:%M:%S', style='{')
ch.setFormatter(formatter)
logger.addHandler(ch)



"""
This module contains helper functions which dispatch
the creation of appropriate generators depending on
the input types.
"""

__all__ = ['SelectOne', 'SelectMultiple']

from .base import TohuBaseGenerator
from .primitive_generators import as_tohu_generator, SelectOnePrimitive, SelectMultiplePrimitive
from .derived_generators import SelectOneDerived, SelectMultipleDerived


def SelectOne(values, p=None):
    if not isinstance(values, TohuBaseGenerator) and not isinstance(p, TohuBaseGenerator):
        return SelectOnePrimitive(values, p)
    else:
        values = as_tohu_generator(values)
        p = as_tohu_generator(p)
        return SelectOneDerived(values, p)


def SelectMultiple(values, num):
    if not isinstance(values, TohuBaseGenerator) and not isinstance(num, TohuBaseGenerator):
        return SelectMultiplePrimitive(values, num)
    else:
        values = as_tohu_generator(values)
        num = as_tohu_generator(num)
        return SelectMultipleDerived(values, num)




from collections import namedtuple

__all__ = ['identity', 'print_generated_sequence']


def identity(x):
    """
    Helper function which returns its argument unchanged.
    That is, `identity(x)` returns `x` for any input `x`.
    """
    return x


def print_generated_sequence(gen, num, *, sep=", ", fmt='', seed=None):
    """
    Helper function which prints a sequence of `num` items
    produced by the random generator `gen`.
    """
    if seed:
       gen.reset(seed)

    elems = [format(next(gen), fmt) for _ in range(num)]
    sep_initial = "\n\n" if '\n' in sep else " "
    print("Generated sequence:{}{}".format(sep_initial, sep.join(elems)))


def make_dummy_tuples(chars='abcde'):
    Quux = namedtuple('Quux', ['x', 'y'])
    some_tuples = [Quux((c*2).upper(), c*2) for c in chars]
    return some_tuples



from .logging import logger


class SpawnMapping:
    def __init__(self):
        self.mapping = dict()

    def __repr__(self):
        return f"<SpawnMapping: {self.mapping}>"

    def __contains__(self, g):
        return g in self.mapping.keys()

    def __setitem__(self, g, g_spawned):
        self.mapping[g] = g_spawned

    def __getitem__(self, g):
        try:
            return self.mapping[g]
        except KeyError:
            logger.debug(f"Generator does not occur in spawn mapping: {g}. Returning it unchanged.")
            return g


import attr
import datetime as dt
import geojson
import numpy as np
import shapely

from faker import Faker
from random import Random
from shapely.geometry import Point, Polygon, MultiPolygon

from .base import TohuBaseGenerator, PrimitiveGenerator, SeedGenerator
from .logging import logger
from .utils import ensure_is_date_object, ensure_is_datetime_object, identity, make_timestamp_formatter, TohuDateError, TohuTimestampError

__all__ = ['Boolean', 'CharString', 'Constant', 'Date', 'DigitString', 'FakerGenerator', 'Float', 'GeoJSONGeolocation',
           'HashDigest', 'Incremental', 'Integer', 'NumpyRandomGenerator', 'Sequential', 'Timestamp', 'as_tohu_generator']


class Constant(PrimitiveGenerator):
    """
    Generator which produces a constant sequence (repeating the same value indefinitely).
    """

    def __init__(self, value):
        """
        Parameters
        ----------
        value:
            The constant value produced by this generator.
        """
        super().__init__()
        self.value = value

    @property
    def max_value(self):
        try:
            self.value < self.value
        except TypeError:
             logger.warning(
                 f"The value '{self.value}' of this Constant generator does not support comparison but"
                 "is being asked for its 'max_value'. There may be an error in the logic somewhere."
            )

        return self.value

    def reset(self, seed=None):
        """
        Note that this method supports the `seed` argument (for consistency with other generators),
        but its value is ignored because resetting a Constant generator has no effect.
        """
        super().reset(seed)
        return self

    def __next__(self):
        return self.value

    def spawn(self, spawn_mapping=None):
        return Constant(self.value)

    def _set_random_state_from(self, other):
        pass


class Boolean(PrimitiveGenerator):
    """
    Generator which produces random boolean values (True or False) with a given probability.
    """

    def __init__(self, p=0.5):
        """
        Parameters
        ----------
        p: float
            The probability that True is returned. Must be between 0.0 and 1.0.
        """
        super().__init__()
        self.p = p
        self.randgen = Random()

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)
        return self

    def __next__(self):
        return self.randgen.random() < self.p

    def spawn(self, spawn_mapping=None):
        new_obj = Boolean(self.p)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.randgen.setstate(other.randgen.getstate())


class Incremental(PrimitiveGenerator):
    """
    Generator which produces integers that increase in regular steps.
    """

    def __init__(self, *, start=0, step=1):
        """
        Parameters
        ----------
        start : int
            Start value of the sequence.
        step : int
            Step size of the sequence.

        Example
        -------
        >>> g = Incremental(start=200, step=4)
        >>> list(g.generate(num=10))
        [200, 204, 208, 212, 216, 220, 224, 228, 232, 236]
        """
        super().__init__()
        self.start = start
        self.step = step
        self.cur_value = start

    def __next__(self):
        retval = self.cur_value
        self.cur_value += self.step
        return retval

    def reset(self, seed=None):
        super().reset(seed)
        self.cur_value = self.start
        return self

    def spawn(self, spawn_mapping=None):
        new_obj = Incremental(start=self.start, step=self.step)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.start = other.start
        self.cur_value = other.cur_value


class Integer(PrimitiveGenerator):
    """
    Generator which produces random integers k in the range low <= k <= high.
    """

    def __init__(self, low, high):
        """
        Parameters
        ----------
        low: integer
            Lower bound (inclusive).
        high: integer
            Upper bound (inclusive).
        """
        super().__init__()
        self.low = low
        self.high = high
        self.randgen = Random()

    @property
    def max_value(self):
        return self.high

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(next(self.seed_generator))
        return self

    def __next__(self):
        return self.randgen.randint(self.low, self.high)

    def spawn(self, spawn_mapping=None):
        new_obj = Integer(self.low, self.high)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.randgen.setstate(other.randgen.getstate())


class Float(PrimitiveGenerator):
    """
    Generator which produces random floating point numbers x in the range low <= x <= high.
    """

    def __init__(self, low, high):
        """
        Parameters
        ----------
        low: integer
            Lower bound (inclusive).
        high: integer
            Upper bound (inclusive).
        """
        super().__init__()
        self.low = low
        self.high = high
        self.randgen = Random()

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)
        return self

    def __next__(self):
        return self.randgen.uniform(self.low, self.high)

    def spawn(self, spawn_mapping=None):
        new_obj = Float(self.low, self.high)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.randgen.setstate(other.randgen.getstate())



CHARACTER_SETS = {
    '<alphanumeric>': 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789',
    '<alphanumeric_lowercase>': 'abcdefghijklmnopqrstuvwxyz0123456789',
    '<alphanumeric_uppercase>': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789',
    '<lowercase>': 'abcdefghijklmnopqrstuvwxyz',
    '<uppercase>': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
    '<digits>': '0123456789',
}


class CharString(PrimitiveGenerator):
    """
    Generator which produces a sequence of character strings.
    """

    def __init__(self, *, length, charset='<alphanumeric>'):
        """
        Parameters
        ----------
        length: integer
            Length of the character strings produced by this generator.
        charset: iterable
            Character set to draw from when generating strings, or string
            with the name of a pre-defined character set.
            Default: <alphanumeric> (both lowercase and uppercase letters).
        """
        super().__init__()
        self.length = length
        try:
            self.charset = CHARACTER_SETS[charset]
            logger.debug(f"Using pre-defined character set: '{charset}'")
        except KeyError:
            self.charset = charset
        self.seed_generator = SeedGenerator()
        self.char_gen = Random()

    def spawn(self, spawn_mapping=None):
        new_obj = CharString(length=self.length, charset=self.charset)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.seed_generator._set_random_state_from(other.seed_generator)
        self.char_gen.setstate(other.char_gen.getstate())

    def __next__(self):
        chars = self.char_gen.choices(self.charset, k=self.length)
        return ''.join(chars)

    def reset(self, seed):
        super().reset(seed)
        self.char_gen.seed(next(self.seed_generator))
        return self


class DigitString(CharString):
    """
    Generator which produces a sequence of strings containing only digits.
    """

    def __init__(self, *, length=None):
        """
        Parameters
        ----------
        length: integer
            Length of the character strings produced by this generator.
        """
        charset = "0123456789"
        super().__init__(length=length, charset=charset)

    def spawn(self, spawn_mapping=None):
        new_obj = DigitString(length=self.length)
        new_obj._set_random_state_from(self)
        return new_obj


class HashDigest(PrimitiveGenerator):
    """
    Generator which produces a sequence of hex strings representing hash digest values.
    """

    def __init__(self, *, length=None, as_bytes=False, uppercase=True):
        """
        Parameters
        ----------
        length: integer
            Length of the character strings produced by this generator.
        as_bytes: bool
            If True, return `length` random bytes. If False, return a string of `length`
            characters with a hexadecimal representation of `length/2` random bytes.
            Note that in the second case `length` must be an even number.
        uppercase: bool
            If True (the default), return hex string using uppercase letters, otherwise lowercase.
            This only has an effect if `as_bytes=False`.
        """
        super().__init__()
        self.length = length
        self._internal_length = length if as_bytes else length / 2
        if not as_bytes and (length % 2) != 0:
            raise ValueError(
                f"Length must be an even number if as_bytes=False because it "
                f"represents length = 2 * num_random_bytes. Got: length={length})")
        self.as_bytes = as_bytes
        self.uppercase = uppercase
        self.randgen = np.random.RandomState()
        self._maybe_convert_to_hex = identity if self.as_bytes else bytes.hex
        self._maybe_convert_to_uppercase = identity if (self.as_bytes or not uppercase) else str.upper

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)
        return self

    def __next__(self):
        val = self.randgen.bytes(self._internal_length)
        return self._maybe_convert_to_uppercase(self._maybe_convert_to_hex(val))

    def spawn(self, spawn_mapping=None):
        new_obj = HashDigest(length=self.length, as_bytes=self.as_bytes, uppercase=self.uppercase)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.randgen.set_state(other.randgen.get_state())


class Sequential(PrimitiveGenerator):
    """
    Generator which produces a sequence of strings
    of the form:

        "PREFIX001"
        "PREFIX002"
        "PREFIX003"
        ...

    Both the prefix and the number of digits can
    be modified by the user.

    Example:
        >>> s = Sequential(prefix="Foobar_", digits=4)
        >>> next(s)
        Foobar_0001
        >>> next(s)
        Foobar_0002
        >>> next(s)
        Foobar_0003
    """

    def __init__(self, *, prefix, digits):
        """
        Parameters
        ----------
        prefix: string
            Prefix to be appended to generated elements.
        digits: integer
            Number of digits to use for the sequential numbering.
            Any numbers will fewer digits will be zero-padded;
            numbers with more digits are unaffected.
        """
        super().__init__()
        self.prefix = prefix
        self.digits = digits
        self.fmt_str = self.prefix + '{{:0{digits}}}'.format(digits=digits)
        self.cnt = 0

    def spawn(self, spawn_mapping=None):
        new_obj = Sequential(prefix=self.prefix, digits=self.digits)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.cnt = other.cnt

    def reset(self, seed=None):
        """
        Note that this method supports the `seed` argument (for consistency with other generators),
        but its value is ignored - the generator is simply reset to its initial value.
        """
        super().reset(seed)
        self.cnt = 0
        return self

    def __next__(self):
        self.cnt += 1
        return self.fmt_str.format(self.cnt)


class NumpyRandomGenerator(TohuBaseGenerator):
    """
    Generator which produces random numbers using one of the methods supported by numpy. [1]

    [1] https://docs.scipy.org/doc/numpy/reference/routines.random.html
    """

    def __init__(self, method, **numpy_args):
        """
        Parameters
        ----------
        method: string
            Name of the numpy function to use (see [1] for details)
        numpy_args:
            Remaining arguments passed to the numpy function (see [1] for details)

        References
        ----------
        [1] https://docs.scipy.org/doc/numpy/reference/routines.random.html
        """
        super().__init__()
        self.method = method
        self.random_state = np.random.RandomState()
        self.randgen = getattr(self.random_state, method)
        self.numpy_args = numpy_args

    def reset(self, seed):
        super().reset(seed)
        self.random_state.seed(seed)
        return self

    def __next__(self):
        return self.randgen(**self.numpy_args)

    def spawn(self, spawn_mapping=None):
        new_obj = NumpyRandomGenerator(method=self.method, **self.numpy_args)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.random_state.set_state(other.random_state.get_state())


class FakerGenerator(PrimitiveGenerator):
    """
    Generator which produces random elements using one of the methods supported by faker. [1]

    [1] https://faker.readthedocs.io/
    """

    def __init__(self, method, *, locale=None, **faker_args):
        """
        Parameters
        ----------
        method: string
            Name of the faker provider to use (see [1] for details)
        locale: string
             Locale to use when generating data, e.g. 'en_US' (see [1] for details)
        faker_args:
            Remaining arguments passed to the faker provider (see [1] for details)

        References
        ----------
        [1] https://faker.readthedocs.io/
        """
        super().__init__()
        self.method = method
        self.locale = locale
        self.faker_args = faker_args

        self.fake = Faker(locale=locale)
        self.randgen = getattr(self.fake, method)
        self.fake.seed_instance(None)  # seed instance to ensure we are decoupled from the global random state

    def reset(self, seed):
        super().reset(seed)
        self.fake.seed_instance(seed)
        return self

    def __next__(self):
        return self.randgen(**self.faker_args)

    def spawn(self, spawn_mapping=None):
        new_obj = FakerGenerator(self.method, locale=self.locale, **self.faker_args)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.fake.random.setstate(other.fake.random.getstate())


class ShapelyGeolocation(PrimitiveGenerator):
    """
    Generator which produces random locations inside a shapely polygon
    or multipolygon. This is a helper class and most users will probably
    find the GeoJSONGeolocation generator more useful.
    """

    def __init__(self, shp, properties=None, max_tries=100):
        if not isinstance(shp, (Polygon, MultiPolygon)):
            raise TypeError(f"Argument 'shp' must be of type Polygon or MultiPolygon. Got: {type(shp)}")

        super().__init__()

        self.shape = shapely.geometry.shape(shp)
        self.properties = properties or dict()

        self.geolocation_cls = self._make_geolocation_class()

        lon_min, lat_min, lon_max, lat_max = self.shape.bounds
        self.lon_gen = Float(lon_min, lon_max)
        self.lat_gen = Float(lat_min, lat_max)
        self.max_tries = max_tries
        self.seed_generator = SeedGenerator()

    def _make_geolocation_class(self):
        fields = {'lon': attr.ib(), 'lat': attr.ib()}
        fields.update({name: attr.ib(value) for name, value in self.properties.items()})
        cls = attr.make_class('Geolocation', fields)
        cls.as_dict = lambda self: attr.asdict(self)
        def __new_eq__(self, other):
            return self.lon == other.lon and self.lat == other.lat
        cls.__eq__ = __new_eq__
        return cls

    def __repr__(self):
        return f"<ShapelyShape, area={self.area:.3f}>"

    def spawn(self, spawn_mapping=None):
        new_obj = ShapelyGeolocation(self.shape, properties=self.properties, max_tries=self.max_tries)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.seed_generator._set_random_state_from(other.seed_generator)
        self.lon_gen._set_random_state_from(other.lon_gen)
        self.lat_gen._set_random_state_from(other.lat_gen)

    @property
    def area(self):
        return self.shape.area

    def __next__(self):
        for cnt in range(1, self.max_tries + 1):
            pt = Point(next(self.lon_gen), next(self.lat_gen))
            if pt.within(self.shape):
                return self.geolocation_cls(lon=pt.x, lat=pt.y)
            else:
                logger.debug(f"Generated point is not within shape. Trying again... [{cnt}/{self.max_tries}]")
        raise RuntimeError(f"Could not generate point in shape after {self.max_tries} attempts")

    def reset(self, seed):
        super().reset(seed)
        self.lon_gen.reset(next(self.seed_generator))
        self.lat_gen.reset(next(self.seed_generator))
        return self


class GeoJSONGeolocation(PrimitiveGenerator):
    """
    Generator which produces random locations inside a geographic area.
    """

    def __init__(self, filename_or_geojson_data, include_attributes=None, max_tries=100):
        super().__init__()

        if isinstance(filename_or_geojson_data, str):
            try:
                with open(filename_or_geojson_data, 'r') as f:
                    geojson_data = geojson.load(f)
            except AttributeError:
                raise NotImplementedError()
        else:
            geojson_data = filename_or_geojson_data

        self.geojson_data = geojson_data
        self.include_attributes = include_attributes or []
        self.max_tries = max_tries

        self.shape_gens = self._make_shape_generators()

        areas = np.array([s.area for s in self.shape_gens])
        self.choice_probs = areas / areas.sum()  # TODO: allow weighin by an arbitrary attribute, not just by area

        self.seed_generator = SeedGenerator()
        self.shape_gen_chooser = np.random.RandomState()

    def _make_shape_generators(self):
        shape_gens = []

        for feature in self.geojson_data['features']:

            geom = shapely.geometry.shape(feature['geometry'])

            cur_attributes = {}
            for name in self.include_attributes:
                try:
                    cur_attributes[name] = feature['properties'][name]
                except KeyError:
                    valid_attributes = list(feature['properties'].keys())
                    raise ValueError(f"Feature does not have attribute '{name}'. Valid attributes are: {valid_attributes}")

            shape_gens.append(ShapelyGeolocation(geom, cur_attributes, max_tries=self.max_tries))

        return shape_gens

    def spawn(self, spawn_mapping=None):
        new_obj = GeoJSONGeolocation(self.geojson_data, include_attributes=self.include_attributes, max_tries=self.max_tries)
        new_obj._set_random_state_from(self)
        return new_obj

    def __next__(self):
        sg = self.shape_gen_chooser.choice(self.shape_gens, p=self.choice_probs)
        return next(sg)

    def reset(self, seed):
        super().reset(seed)
        self.shape_gen_chooser.seed(next(self.seed_generator))
        for g in self.shape_gens:
            g.reset(next(self.seed_generator))
        return self

    def _set_random_state_from(self, other):
        self.seed_generator._set_random_state_from(other.seed_generator)
        self.shape_gen_chooser.set_state(other.shape_gen_chooser.get_state())
        for gen_self, gen_other in zip(self.shape_gens, other.shape_gens):
            gen_self._set_random_state_from(gen_other)


def as_tohu_generator(g):
    """
    Helper function to ensure that a given input is a tohu generator.

    If g is already a tohu generator then it is returned unchanged,
    otherwise it is wrapped in a Constant generator.
    """

    if isinstance(g, TohuBaseGenerator):
        return g
    else:
        return Constant(g)


def get_start_and_end_values(start, end, date):
    if start is not None and end is not None and date is not None:
        raise TohuTimestampError("Arguments 'start', 'end', 'date' must not all be given.")

    if date is None and (start is None or end is None):
        raise TohuTimestampError("If argument 'date' is not given, both 'start' and 'end' must be provided.")

    if date is not None:
        date = ensure_is_date_object(date)

    if start is None:
        start = ensure_is_datetime_object(date)
    else:
        start = ensure_is_datetime_object(start)

    if end is None:
        end = ensure_is_datetime_object(date, optional_offset=dt.timedelta(hours=23, minutes=59, seconds=59))
    else:
        end = ensure_is_datetime_object(end)

    if date is not None:
        if start.date() != date:
            raise TohuTimestampError("Date of start timestamp does not coincide with date argument.")
        if end.date() != date:
            raise TohuTimestampError("Date of end timestamp does not coincide with date argument.")


    return start, end


class Timestamp(TohuBaseGenerator):

    def __init__(self, *, start=None, end=None, date=None, fmt=None, uppercase=None):
        super().__init__()
        self.start, self.end = get_start_and_end_values(start, end, date)
        self.interval = (self.end - self.start).total_seconds()
        self.offset_randgen = Random()
        self._check_start_before_end()

        self.fmt = fmt
        self.uppercase = uppercase
        self._maybe_format_timestamp = make_timestamp_formatter(self.fmt, self.uppercase)

    @property
    def max_value(self):
        return self.end

    def _check_start_before_end(self):
        if self.start > self.end:
            raise TohuTimestampError(f"Start value must be before end value. Got: start={self.start}, end={self.end}")

    def __next__(self):
        offset = self.offset_randgen.randint(0, self.interval)
        ts = self.start + dt.timedelta(seconds=offset)
        return self._maybe_format_timestamp(ts)

    def reset(self, seed):
        super().reset(seed)
        self.offset_randgen.seed(next(self.seed_generator))
        return self

    def spawn(self, spawn_mapping=None):
        new_obj = Timestamp(start=self.start, end=self.end, fmt=self.fmt, uppercase=self.uppercase)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.offset_randgen.setstate(other.offset_randgen.getstate())

    def strftime(self, fmt='%Y-%m-%d %H:%M:%S', uppercase=False):
        g = Timestamp(start=self.start, end=self.end, fmt=fmt, uppercase=uppercase)
        self.register_clone(g)
        g.register_parent(self)
        return g


class Date(TohuBaseGenerator):

    def __init__(self, start, end, *, fmt=None, uppercase=None):
        super().__init__()
        self.start = ensure_is_date_object(start)
        self.end = ensure_is_date_object(end)
        self.interval = (self.end - self.start).days
        self.offset_randgen = Random()
        self._check_start_before_end()

        self.fmt = fmt
        self.uppercase = uppercase
        self._maybe_format_timestamp = make_timestamp_formatter(self.fmt, self.uppercase)

    def _check_start_before_end(self):
        if self.start > self.end:
            raise TohuDateError(f"Start value must be before end value. Got: start={self.start}, end={self.end}")

    def __next__(self):
        offset = self.offset_randgen.randint(0, self.interval)
        ds = self.start + dt.timedelta(days=offset)
        return self._maybe_format_timestamp(ds)

    def reset(self, seed):
        super().reset(seed)
        self.offset_randgen.seed(next(self.seed_generator))
        return self

    def spawn(self, spawn_mapping=None):
        new_obj = Date(self.start, self.end, fmt=self.fmt, uppercase=self.uppercase)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.offset_randgen.setstate(other.offset_randgen.getstate())

    def strftime(self, fmt='%Y-%m-%d', uppercase=False):
        g = Timestamp(start=self.start, end=self.end, fmt=fmt, uppercase=uppercase)
        self.register_clone(g)
        g.register_parent(self)
        return g



import datetime as dt

from operator import attrgetter
from random import Random

from .base import TohuBaseGenerator, SeedGenerator
from .logging import logger
from .primitive_generators import as_tohu_generator, Constant, Date, Timestamp as TimestampPrimitive
from .spawn_mapping import SpawnMapping
from .utils import TohuDateError, TohuTimestampError, ensure_is_date_object, make_timestamp_formatter

__all__ = ['Apply', 'Cumsum', 'GetAttribute', 'Integer', 'Lookup', 'MultiCumsum', 'SelectMultiple', 'SelectOne', 'Tee', 'Timestamp']


class DerivedGenerator(TohuBaseGenerator):
    """
    Base class for all derived generators
    """

    def reset_input_generators(self, seed):
        """
        Helper method which explicitly resets all input generators
        to the derived generator. This should only ever be called
        for testing or debugging.
        """
        seed_generator = SeedGenerator().reset(seed=seed)

        for gen in self.input_generators:
            gen.reset(next(seed_generator))
            try:
                # In case `gen` is itself a derived generator,
                # recursively reset its own input generators.
                gen.reset_input_generators(next(seed_generator))
            except AttributeError:
                pass


class Apply(DerivedGenerator):
    """
    Generator which applies a callable to a elements produced by a set of input generators.
    """

    def __init__(self, callable, *arg_gens, max_value=None, **kwarg_gens):
        super().__init__()
        self.callable = callable
        self.arg_gens_orig = arg_gens
        self.kwarg_gens_orig = kwarg_gens
        self.max_value = max_value

        self.arg_gens = [g.clone() for g in self.arg_gens_orig]
        self.kwarg_gens = {name: g.clone() for name, g in self.kwarg_gens_orig.items()}
        self.input_generators = [g for g in self.arg_gens_orig] + [g for g in self.kwarg_gens_orig.values()]
        self.constituent_generators = [g for g in self.arg_gens] + [g for g in self.kwarg_gens.values()]
        for gen in self.constituent_generators:
            gen.owner = self

    def __next__(self):
        args = [next(g) for g in self.arg_gens]
        kwargs = {name: next(g) for name, g in self.kwarg_gens.items()}
        return self.callable(*args, **kwargs)

    def reset(self, seed):
        super().reset(seed)

        try:
            self.callable.reset(next(self.seed_generator))
        except AttributeError:
            logger.debug(
                f"Failed to reset callable in generator {self}. Assuming that "
                "it does not contain any random generators that need resetting."
            )

        return self

    def spawn(self, spawn_mapping=None):
        spawn_mapping = spawn_mapping or SpawnMapping()
        new_arg_gens_orig = [spawn_mapping[g] for g in self.arg_gens_orig]
        new_kwarg_gens_orig = {name: spawn_mapping[g] for name, g in self.kwarg_gens_orig}
        new_obj = Apply(self.callable, *new_arg_gens_orig, max_value=self.max_value, **new_kwarg_gens_orig)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        for g_self, g_other in zip(self.constituent_generators, other.constituent_generators):
            g_self._set_random_state_from(g_other)


class GetAttribute(Apply):

    def __init__(self, g, name):
        self.g = as_tohu_generator(g)  # no need to clone here because this happens in the superclass
        self.name = as_tohu_generator(name)

        def func(value, name):
            # TODO: this is not very efficient if `name` is a constant
            # string because we're re-creating the attrgetter every time.
            # However, since `name` can in principle be a generator itself
            # we need to keep the generality.
            f = attrgetter(name)

            try:
                return f(value)
            except AttributeError:
                try:
                    return [f(x) for x in value]
                except TypeError:
                    raise AttributeError(f"Could not extract attribute '{name}' from {value}")
                except AttributeError:
                    raise AttributeError(f"Could not extract attribute '{name}' from items in sequence: {value}")

        super().__init__(func, self.g, self.name)

    def spawn(self, spawn_mapping=None):
        spawn_mapping = spawn_mapping or SpawnMapping()
        new_obj = GetAttribute(spawn_mapping[self.g], spawn_mapping[self.name])
        new_obj._set_random_state_from(self)
        return new_obj

    def reset(self, seed):
        super().reset(seed)
        self.g.reset(seed)
        return self


class Lookup(Apply):
    """
    Generator which performs a lookup of elements produced by another generator.
    """

    def __init__(self, key, mapping):
        self.key = as_tohu_generator(key)
        self.mapping = as_tohu_generator(mapping)

        def f_lookup(key, mapping):
            return mapping[key]

        super().__init__(f_lookup, self.key, self.mapping)

    def spawn(self, spawn_mapping=None):
        spawn_mapping = spawn_mapping or SpawnMapping()
        new_obj = Lookup(spawn_mapping[self.key], spawn_mapping[self.mapping])
        new_obj._set_random_state_from(self)
        return new_obj


class Integer(Apply):

    def __init__(self, low, high):
        self.low_gen = as_tohu_generator(low)
        self.high_gen = as_tohu_generator(high)
        self.randgen = Random()
        super().__init__(self.randgen.randint, self.low_gen, self.high_gen)
        self.max_value = self.high_gen.max_value

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(next(self.seed_generator))
        return self

    def spawn(self, spawn_mapping=None):
        spawn_mapping = spawn_mapping or SpawnMapping()
        new_obj = Integer(spawn_mapping[self.low_gen], spawn_mapping[self.high_gen])
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.randgen.setstate(other.randgen.getstate())


class SelectOne(Apply):
    """
    Generator which selects a single element from each sequence produced by another generator.
    """

    def __init__(self, values, p=None):
        self.values_gen = as_tohu_generator(values)
        self.p_gen = as_tohu_generator(p)
        self.randgen = Random()

        def func(values, p):
            return self.randgen.choices(values, weights=p)[0]

        super().__init__(func, self.values_gen, self.p_gen)

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)
        return self

    def spawn(self, spawn_mapping=None):
        spawn_mapping = spawn_mapping or SpawnMapping()
        new_obj = SelectOne(spawn_mapping[self.values_gen], p=spawn_mapping[self.p_gen])
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.randgen.setstate(other.randgen.getstate())

    def _spot_check_that_elements_produced_by_this_generator_have_attribute(self, name):
        """
        Helper function to spot-check that the items produces by this generator have the attribute `name`.
        """
        g_tmp = self.values_gen.spawn()
        sample_element = next(g_tmp)[0]
        try:
            getattr(sample_element, name)
        except AttributeError:
            raise AttributeError(f"Items produced by {self} do not have the attribute '{name}'")

    def __getattr__(self, name):
        self._spot_check_that_elements_produced_by_this_generator_have_attribute(name)
        return GetAttribute(self, name)



class SelectMultiple(Apply):
    """
    Generator which selects multiple elements (without replacement)
    from each sequence produced by another generator.
    """

    def __init__(self, values, num):
        self.values_gen = as_tohu_generator(values)
        self.num_gen = as_tohu_generator(num)
        self.randgen = Random()
        func = self.randgen.sample
        super().__init__(func, self.values_gen, k=self.num_gen)

    def reset(self, seed):
        super().reset(seed)
        self.randgen.seed(seed)
        return self

    def spawn(self, spawn_mapping=None):
        spawn_mapping = spawn_mapping or SpawnMapping()
        new_obj = SelectMultiple(spawn_mapping[self.values_gen], spawn_mapping[self.num_gen])
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.randgen.setstate(other.randgen.getstate())

    def size(self):
        def get_size(x):
            return len(x)
        g = Apply(get_size, self, max_value=self.num_gen.max_value)
        return g


class Tee(Apply):

    def __init__(self, g, num):
        self.g_orig = g
        self.num_gen = as_tohu_generator(num)
        self.value_gens = [g.spawn() for _ in range(self.num_gen.max_value)]

        if self.num_gen.max_value > 1000:
            raise NotImplementedError(
                "This Tee generator is intended to be used to produce small-ish output tuples. "
                "The current implementation is not ideal for potentially large tuples, which"
                "which is why we only allow sizes up to 1000 elements at the moment."
            )

        def make_tuple(num, *values):
            return tuple(values[:num])

        super().__init__(make_tuple, self.num_gen, *self.value_gens)

    def reset(self, seed):
        super().reset(seed)

        # We need to explicitly reset the value generators because they
        # are not technically input generators to this derived generator
        # so they can't be externally reset.
        for g in self.value_gens:
            g.reset(next(self.seed_generator))

        return self


def convert_to_date_object(date):
    if isinstance(date, Constant):
        return convert_to_date_object(date.value)
    elif isinstance(date, Date) and date.start == date.end:
        return date.start
    else:
        try:
            return ensure_is_date_object(date)
        except TohuDateError:
            raise TohuTimestampError(f"Argument 'date' must represent some kind of constant date object. Got: {date}")


def get_start_generator(start, date):
    if date is not None:
        date = convert_to_date_object(date)

    if start is None:
        start_value = dt.datetime(date.year, date.month, date.day)
        start_gen = Constant(start_value)
    elif isinstance(start, str):
        start_value = dt.datetime.strptime(start, "%Y-%m-%d %H:%M:%S")
        start_gen = Constant(start_value)
    elif isinstance(start, dt.datetime):
        return Constant(start)
    elif isinstance(start, Constant):
        return get_start_generator(start.value, date)
    elif isinstance(start, TimestampPrimitive):
        # Create a new generator to strip any string formatting information in case it exists
        start_without_formatting = TimestampPrimitive(start=start.start, end=start.end)
        start.register_clone(start_without_formatting)
        start_without_formatting.register_parent(start)
        return start_without_formatting
    else:
        raise NotImplementedError()

    return start_gen


def get_end_generator(end, date):
    if date is not None:
        date = convert_to_date_object(date)

    if end is None:
        end_value = dt.datetime(date.year, date.month, date.day, 23, 59, 59)
        end_gen = Constant(end_value)
    elif isinstance(end, str):
        end_value = dt.datetime.strptime(end, "%Y-%m-%d %H:%M:%S")
        end_gen = Constant(end_value)
    elif isinstance(end, dt.datetime):
        return Constant(end)
    elif isinstance(end, Constant):
        return get_start_generator(end.value, date)
    elif isinstance(end, TimestampPrimitive):
        # Create a new generator to strip any string formatting information in case it exists
        end_without_formatting = TimestampPrimitive(start=end.start, end=end.end)
        end.register_clone(end_without_formatting)
        end_without_formatting.register_parent(end)
        return end_without_formatting
    else:
        raise NotImplementedError()

    return end_gen


def get_start_end_end_generator(start, end, date):
    start_gen = get_start_generator(start, date)
    end_gen = get_end_generator(end, date)
    return start_gen, end_gen


def check_valid_inputs(start_gen, end_gen, date):
    if date is not None:
        date = convert_to_date_object(date)

    if date is not None:
        if isinstance(start_gen, TimestampPrimitive):
            if not (start_gen.start.date() == date and start_gen.end.date() == date):
                raise TohuTimestampError(
                    "If the 'date' argument is given, all possible 'start' timestamp values must lie on that given date."
                )
        if isinstance(end_gen, TimestampPrimitive):
            if not (end_gen.start.date() == date and end_gen.end.date() == date):
                raise TohuTimestampError(
                    "If the 'date' argument is given, all possible 'end' timestamp values must lie on that given date."
                )

    start_end_error_msg = (
        "Latest possible value of 'start' generator must not be after "
        "earliest possible value of 'end' generator."
    )

    if isinstance(start_gen, TimestampPrimitive) and isinstance(end_gen, TimestampPrimitive):
        if start_gen.end > end_gen.start:
            raise TohuTimestampError(start_end_error_msg)
    elif isinstance(start_gen, TimestampPrimitive) and isinstance(end_gen, Constant):
        if start_gen.end > end_gen.value:
            raise TohuTimestampError(start_end_error_msg)
    elif isinstance(start_gen, Constant) and isinstance(end_gen, TimestampPrimitive):
        if start_gen.value > end_gen.start:
            raise TohuTimestampError(start_end_error_msg)
    elif isinstance(start_gen, Constant) and isinstance(end_gen, Constant):
        if start_gen.value> end_gen.value:
            raise TohuTimestampError("Start value must be before end value. Got: start={self.start}, end={self.end}")


class Timestamp(Apply):

    def __init__(self, *, start=None, end=None, date=None, fmt=None, uppercase=None):

        if start is None and end is None and date is None:
            raise TohuTimestampError("Not all input arguments can be None.")

        if start is not None and end is not None and date is not None:
            raise TohuTimestampError("Arguments 'start', 'end', 'date' cannot all be provided.")

        self.start_gen, self.end_gen = get_start_end_end_generator(start, end, date)
        check_valid_inputs(self.start_gen, self.end_gen, date)

        self.offset_randgen = Random()

        def func(start, end):
            interval = (end - start).total_seconds()
            try:
                offset = self.offset_randgen.randint(0, interval)
            except ValueError:
                raise TohuTimestampError(f"Start generator produced timestamp later than end generator: start={start}, end={end}")
            ts = (start + dt.timedelta(seconds=offset))
            return ts

        super().__init__(func, self.start_gen, self.end_gen)

        self.max_value = self.end_gen.max_value

        self.fmt = fmt
        self.uppercase = uppercase
        self._maybe_format_timestamp = make_timestamp_formatter(self.fmt, self.uppercase)

    def __next__(self):
        ts = super().__next__()
        return self._maybe_format_timestamp(ts)

    def reset(self, seed):
        super().reset(seed)
        self.offset_randgen.seed(next(self.seed_generator))
        return self

    def spawn(self, spawn_mapping=None):
        spawn_mapping = spawn_mapping or SpawnMapping()
        new_obj = Timestamp(start=spawn_mapping[self.start_gen], end=spawn_mapping[self.end_gen], fmt=self.fmt, uppercase=self.uppercase)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.offset_randgen.setstate(other.offset_randgen.getstate())

    def strftime(self, fmt='%Y-%m-%d %H:%M:%S', uppercase=False):
        g = Timestamp(start=self.start_gen, end=self.end_gen, fmt=fmt, uppercase=uppercase)
        self.register_clone(g)
        g.register_parent(self)
        return g


class Cumsum(DerivedGenerator):

    def __init__(self, g, *, start_with_zero=False):
        super().__init__()
        self.g_orig = g
        self.g_internal = g.clone()
        self.g_internal.owner = self
        self.start_with_zero = start_with_zero
        self.input_generators = [self.g_orig]
        self.constituent_generators = [self.g_internal]
        self.reset()

    def __next__(self):
        retval = self.value
        self.value += next(self.g_internal)
        return retval

    def reset(self, seed=None):
        super().reset(seed)

        if self.start_with_zero:
            self.value = 0
        else:
            # Note: for this to work correctly the input generator `g`
            # needs to be reset _before_ this one.
            self.value = next(self.g_internal)

        return self

    def spawn(self, spawn_mapping=None):
        spawn_mapping = spawn_mapping or SpawnMapping()
        new_obj = Cumsum(spawn_mapping[self.g_orig], start_with_zero=self.start_with_zero)
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)
        self.value = other.value
        self.g_internal._set_random_state_from(other.g_internal)


class MultiCumsum(DerivedGenerator):
    """
    TODO: document me!
    """

    def __init__(self, g, attr_name, g_amount):
        """
        TODO: document me!
        """
        super().__init__()
        self.g_amount_orig = as_tohu_generator(g_amount)
        self.g_amount_internal = g_amount.clone()
        self.g_amount_internal.owner = self

        self.g_orig = g
        self.g_internal = g.clone()
        self.g_internal.owner = self

        self.attr_name = attr_name

        self.input_generators = [self.g_amount_orig, self.g_orig]
        self.constituent_generators = [self.g_amount_internal, self.g_internal]

        self.cur_values = {}

    def __next__(self):
        # TODO: more meaningful variable names!
        x = next(self.g_internal)
        try:
            cur_val = self.cur_values[x]
        except KeyError:
            cur_val = getattr(x, self.attr_name)
            self.cur_values[x] = cur_val

        self.cur_values[x] += next(self.g_amount_internal)
        return cur_val

    def reset(self, seed=None):
        super().reset(seed)
        self.cur_values = {}
        return self

    def spawn(self, spawn_mapping=None):
        spawn_mapping = spawn_mapping or SpawnMapping()
        new_obj = MultiCumsum(spawn_mapping[self.g_orig], self.attr_name, spawn_mapping[self.g_amount_orig])
        new_obj._set_random_state_from(self)
        return new_obj

    def _set_random_state_from(self, other):
        self.g_amount_internal._set_random_state_from(other.g_amount_internal)
        self.g_internal._set_random_state_from(other.g_internal)
        self.cur_values = other.cur_values.copy()  ### XXX TODO: can we simply copy these over in all cases?!


"""
This module contains helper functions which dispatch
the creation of appropriate generators depending on
the input types.
"""

__all__ = ['Integer', 'Timestamp']

from .base import TohuBaseGenerator
from .primitive_generators import Integer as IntegerPrimitive, Timestamp as TimestampPrimitive
from .derived_generators import Integer as IntegerDerived,Timestamp as TimestampDerived


def Integer(low, high):
    if (not isinstance(low, TohuBaseGenerator) and not isinstance(high, TohuBaseGenerator)):
        return IntegerPrimitive(low, high)
    else:
        return IntegerDerived(low, high)


def Timestamp(*, start=None, end=None, date=None, fmt=None, uppercase=False):
    if (not isinstance(start, TohuBaseGenerator)
        and not isinstance(end, TohuBaseGenerator)
        and not isinstance(date, TohuBaseGenerator)):
        return TimestampPrimitive(start=start, end=end, date=date, fmt=fmt, uppercase=uppercase)
    else:
        return TimestampDerived(start=start, end=end, date=date, fmt=fmt, uppercase=uppercase)



from distutils.version import StrictVersion
from platform import python_version

min_supported_python_version = '3.6'

if StrictVersion(python_version()) < StrictVersion(min_supported_python_version):
    error_msg = (
        "Tohu requires Python {min_supported_python_version} or greater to run "
        "(currently running under Python {python_version()})"
    )
    raise RuntimeError(error_msg)

from . import v6

from .v6.base import *
from .v6.primitive_generators import *
from .v6.derived_generators import *
from .v6.generator_dispatch import *
from .v6.custom_generator import CustomGenerator
from .v6.logging import logger
from .v6.utils import print_generated_sequence, print_tohu_version

from .v6 import base
from .v6 import primitive_generators
from .v6 import derived_generators
from .v6 import generator_dispatch
from .v6 import custom_generator
from .v6 import set_special_methods

__all__ = base.__all__ \
          + primitive_generators.__all__ \
          + derived_generators.__all__ \
          + generator_dispatch.__all__ \
          + custom_generator.__all__ \
          + ['tohu_logger', 'print_generated_sequence', 'print_tohu_version']

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

tohu_logger = logger  # alias


from itertools import count

from .base import TohuBaseGenerator, SeedGenerator
from .logging import logger
from .spawn_mapping import SpawnMapping
from .utils import is_clone


class TohuNamespaceError(Exception):
    """
    Custom exception for TohuNamespace
    """


class TohuNamespace:

    _global_count = count(start=1)

    def __init__(self):
        self._ns = {}
        self.seed_generator = SeedGenerator()
        self._idx = next(self._global_count)

    def __repr__(self):
        return f"<TohuNamespace #{self._idx}>"

    @classmethod
    def from_dict(cls, d):
        ns = TohuNamespace()
        ns.update_from_dict(d)
        return ns

    def update_from_dict(self, d):
        for name, g in d.items():
            if isinstance(g, TohuBaseGenerator):
                self[name] = g

    def __len__(self):
        return len(self._ns)

    def __contains__(self, g):
        return g in self._ns.keys()

    def to_str(self):
        s = ""
        for g, name in self._ns.items():
            prefix = "    " if name is None else f"{name}: "
            suffix = " (anonymous)" if name is None else ""
            s += f"{prefix}{g}{suffix}\n"
        return s

    @property
    def all_generators(self):
        """
        Return all generators in this namespace.
        """
        return self._ns

    @property
    def all_independent_generators(self):
        """
        Return all generators in this namespace which are not clones.
        """
        return {g: name for g, name in self._ns.items() if not is_clone(g)}

    @property
    def named_generators(self):
        return {name: g for g, name in self._ns.items() if name is not None}

    @property
    def names(self):
        return [x for x in self._ns.values() if x is not None]

    def __getitem__(self, key):
        for g, name in self._ns.items():
            if name == key:
                return g
        raise KeyError(f"No generator with name '{key}' exists in this namespace.")

    def _add(self, g, name):
        for g_input in g.input_generators:
            self._add(g_input, name=None)

        if g.parent is not None:
            self._add(g.parent, name=None)

        if g not in self._ns:
            logger.debug(f"Adding generator to {self}: {g} (name='{name}')")
            self._ns[g] = name

    def __setitem__(self, name, g):
        assert isinstance(g, TohuBaseGenerator)
        if g in self:
            existing_name = self._ns[g]
            if name == existing_name:
                logger.debug(f"Generator already exists with the same name: {g}. Not adding again.")
            else:
                logger.debug("Trying to add existing generator with a different name. Adding a clone instead.")
                self._ns[g.clone()] = name
        else:
            if name in self._ns.values():
                # TODO: is keeping the existing generator and ignoring the new one the right thing to do in all cases?
                logger.debug(f"A different generator already exists with name '{name}'. Ignoring the new one.")
            else:
                self._add(g, name)

    def spawn_generator(self, g, spawn_mapping, ns_spawned):
        logger.debug(f"Spawning generator in {self}: {g}")
        name = self._ns[g]

        if g in ns_spawned:
            # Generator was spawned before; nothing to do
            return
        else:
            if g.parent is not None:
                # Re-wire the clone
                if g.parent not in spawn_mapping:
                    raise TohuNamespaceError(
                        f"The cloned generator {g} is present in the tohu namespace "
                        f"but its parent {g.parent} is not. This should never happen."
                    )
                g_new = g.clone()
                new_parent = spawn_mapping[g.parent]
                g.unregister_clone(g_new)
                g_new.register_parent(new_parent)
                new_parent.register_clone(g_new)
            else:
                # Sanity check that all input generators of g have already been spawned before.
                for g_input in g.input_generators:
                    if g_input not in spawn_mapping:
                        raise TohuNamespaceError(
                            f"Spawn mapping is missing the input generator {g_input} for generator {g}. "
                            f"This should not happen! Need to check the logic of the implementation."
                        )

                # Simply spawn the generator
                g_new = g.spawn(spawn_mapping)

            spawn_mapping[g] = g_new
            ns_spawned[name] = g_new

            # Copy tohu names for easier debugging (so that
            # a custom generator instance has the same tohu
            # names for constituent generators as the class).
            g_new.set_tohu_name(g.tohu_name)

    def spawn(self):
        spawn_mapping = SpawnMapping()
        ns_spawned = TohuNamespace()
        for g, name in self._ns.items():
            self.spawn_generator(g, spawn_mapping, ns_spawned)
            #ns_spawned[name] = g.spawn()
        return ns_spawned

    def reset(self, seed):
        self.seed_generator.reset(seed)
        for g in self.all_independent_generators:
            g.reset(next(self.seed_generator))

    def _set_random_state_from(self, other):
        # TODO: double-check that this traverses generators in both namespaces in the same order
        for g_self, g_other in zip(self._ns, other._ns):
            g_self._set_random_state_from(g_other)

    def set_owner(self, owner):
        for g in self._ns:
            g.owner = owner



import datetime as dt
import pandas as pd
from collections import namedtuple

from .._version import get_versions

__all__ = ['ensure_is_date_object', 'ensure_is_datetime_object', 'identity', 'make_timestamp_formatter',
           'print_generated_sequence', 'parse_date_string', 'parse_datetime_string', 'print_tohu_version']


def print_tohu_version():
    """
    Convenience helper function to print the current tohu version.
    """
    print(f"Tohu version: {get_versions()['version']}")


def identity(x):
    """
    Helper function which returns its argument unchanged.
    That is, `identity(x)` returns `x` for any input `x`.
    """
    return x


def is_clone(g):
    """
    Return True if the generator `g` is a clone of another generator, otherwise return False.
    """
    return g.parent is not None


def make_timestamp_formatter(fmt, uppercase=None):
    if fmt is None:
        formatter = identity
    else:
        if not isinstance(fmt, str):
            raise ValueError(f"Argument 'fmt' must be of type string, got '{type(fmt)}'")

        if uppercase:
            formatter = lambda ts: ts.strftime(fmt).upper()
        else:
            formatter = lambda ts: ts.strftime(fmt)

    return formatter


def make_exploded_column(df, colname_new, colname_old):
    """
    Internal helper function used by `explode_columns()`.
    """
    s = df[colname_old].apply(pd.Series).stack()
    s.name = colname_new
    return s


def explode_columns(df, colnames):
    """
    Given a dataframe with certain columns that contain lists,
    return another dataframe where the elements in each list
    are "exploded" into individual rows.

    Example:

    >>> df
       col1 col2             col3          col4
    0   foo   11  [DDD, AAA, CCC]  [dd, aa, cc]
    1   bar   22            [FFF]          [ff]
    2  quux   33       [EEE, BBB]      [ee, bb]

    >>> explode_columns(df, ['col3'])
       col1 col2 col3 col4
    0   foo   11  DDD   dd
    1   foo   11  AAA   aa
    2   foo   11  CCC   cc
    3   bar   22  FFF   ff
    4  quux   33  EEE   ee
    5  quux   33  BBB   bb

    >>> explode_columns(df, {'col3_exploded': 'col3'})
       col1 col2 col3_exploded col4
    0   foo   11           DDD   dd
    1   foo   11           AAA   aa
    2   foo   11           CCC   cc
    3   bar   22           FFF   ff
    4  quux   33           EEE   ee
    5  quux   33           BBB   bb
    """
    if isinstance(colnames, (list, tuple)):
        colnames = {name: name for name in colnames}

    remaining_columns = list(df.columns.difference(colnames.values()))
    df2 = df.set_index(remaining_columns)
    df3 = pd.concat((make_exploded_column(df2, col_new, col_old) for col_new, col_old in colnames.items()), axis=1)
    levels_to_reset = list(range(len(remaining_columns)))
    df4 = df3.reset_index(level=levels_to_reset).reset_index(drop=True)
    return df4


def print_generated_sequence(gen, num, *, sep=", ", fmt='', seed=None):
    """
    Helper function which prints a sequence of `num` items
    produced by the random generator `gen`.
    """
    if seed:
       gen.reset(seed)

    elems = [format(next(gen), fmt) for _ in range(num)]
    sep_initial = "\n\n" if '\n' in sep else " "
    print("Generated sequence:{}{}".format(sep_initial, sep.join(elems)))


def make_dummy_tuples(chars='abcde'):
    """
    Helper function to create a list of namedtuples which are useful
    for testing and debugging (especially of custom generators).

    Example
    -------
    >>> make_dummy_tuples(chars='abcd')
    [Quux(x='AA', y='aa'),
     Quux(x='BB', y='bb'),
     Quux(x='CC', y='cc'),
     Quux(x='DD', y='dd')]
    """
    Quux = namedtuple('Quux', ['x', 'y'])
    some_tuples = [Quux((c*2).upper(), c*2) for c in chars]
    return some_tuples


class TohuDateError(Exception):
    """
    Custom exception
    """


def parse_date_string(s):
    try:
        date = dt.datetime.strptime(s, "%Y-%m-%d").date()
    except ValueError:
        raise TohuDateError(
            "If input is a string, it must represent a timestamp of the form 'YYYY-MM-DD HH:MM:SS' "
            f"or a date of the form YYYY-MM-DD. Got: '{s}'"
        )
    return date


def ensure_is_date_object(x):
    """
    Ensure input represents a valid date and return the corresponding `datetime.date` object.

    Valid inputs:

      - string of the form "YYYY-MM-DD"
      - dt.date object
      - pd.Timestamp of the form "YYYY-MM-DD 00:00:00" with freq='D' (as is generated by pd.date_range())
    """
    error_msg = f"Cannot convert input to date object: {x} (type: {type(x)})"

    if isinstance(x, dt.date):
        if isinstance(x, pd.Timestamp):
            if x.freq != 'D':
                raise TohuDateError("Pandas Timestamp must have freq='D' set. Got: freq={x.freq!r}")
            elif pd.Timestamp(x.date()) == x:
                return x.date()
            else:
                raise TohuDateError(error_msg)
        elif isinstance(x, dt.datetime):
            raise TohuDateError(error_msg)
        else:
            return x
    elif isinstance(x, str):
        return parse_date_string(x)
    else:
        raise TohuDateError(error_msg)


class TohuTimestampError(Exception):
    """
    Custom exception
    """


def is_date_object(x):
    return isinstance(x, dt.date) and not isinstance(x, dt.datetime)


def is_date_string(x):
    if not isinstance(x, str):
        return False
    else:
        try:
            dt.datetime.strptime(x, "%Y-%m-%d")
            return True
        except ValueError:
            return False


def parse_datetime_string(s, optional_offset=None):
    try:
        ts = dt.datetime.strptime(s, "%Y-%m-%d %H:%M:%S")
    except ValueError:
        optional_offset = optional_offset or dt.timedelta(seconds=0)
        try:
            ts = dt.datetime.strptime(s, "%Y-%m-%d") + optional_offset
        except ValueError:
            raise ValueError(
                "If input is a string, it must represent a timestamp of the form 'YYYY-MM-DD HH:MM:SS' "
                f"or a date of the form YYYY-MM-DD. Got: '{s}'"
            )
    return ts


def ensure_is_datetime_object(x, optional_offset=None):
    if isinstance(x, dt.datetime):
        return x
    elif isinstance(x, dt.date):
        optional_offset = optional_offset or dt.timedelta(seconds=0)
        return dt.datetime(x.year, x.month, x.day) + optional_offset
    elif isinstance(x, str):
        return parse_datetime_string(x, optional_offset)
    else:
        raise TohuTimestampError(f"Cannot convert input to datetime object: {x} (type: {type(x)})")



from .primitive_generators import *
from .derived_generators import *
from .generator_dispatch import *
from .custom_generator import *
from .logging import logger as tohu_logger
from .utils import print_tohu_version


from .base import TohuBaseGenerator
from .utils import make_tohu_item_class

__all__ = ['CustomGenerator']


class CustomGenerator(TohuBaseGenerator):

    def __init__(self, *args, **kwargs):
        super().__init__()

        self.orig_args = args
        self.orig_kwargs = kwargs

        self._set_tohu_item_class()

    def _set_tohu_item_class(self):
        #self.tohu_item_cls = make_tohu_item_class(self.__tohu_items_name__, self.field_names)
        pass

    def __next__(self):
        # field_values = [next(g) for g in self.field_generators.values()]
        # return self.tohu_item_cls(*field_values)
        return None

    def reset(self, seed):
        super().reset(seed)
        # for gen in self.field_generators.values():
        #     gen.reset(next(self.seed_generator))

    def spawn(self):
        new_obj = self.__class__(*self.orig_args, **self.orig_kwargs)
        new_obj._set_random_state_from(self)

        # # Explicitly set item_cls. This is necessary because due to
        # # the way in which `attr` works, explicit comparisons between
        # # generated items will return False even though they contain
        # # the same elements (because the underlying attr classes are
        # # different, so attr plays it safe).
        # new_obj.tohu_item_cls = self.tohu_item_cls

        return new_obj

    def _set_random_state_from(self, other):
        super()._set_random_state_from(other)

        # # TODO: should also set random state for unnamed field generators
        # #        (these can occur in chains of derived generators)
        # for name in self.field_generators.keys():
        #     self.field_generators[name]._set_random_state_from(other.field_generators[name])



"""
This module is not meant to be imported directly.
Its purpose is to patch the TohuBaseGenerator class
so that its special methods __add__, __mul__ etc.
support other generators as arguments.
"""

from .base import TohuBaseGenerator
from .primitive_generators import GeoJSONGeolocation, as_tohu_generator
from .derived_generators import Apply, GetAttribute
from operator import add, mul, gt, ge, lt, le, eq

__all__ = []


def check_that_operator_can_be_applied_to_produces_items(op, g1, g2):
    """
    Helper function to check that the operator `op` can be applied to items produced by g1 and g2.
    """
    g1_tmp_copy = g1.spawn()
    g2_tmp_copy = g2.spawn()
    sample_item_1 = next(g1_tmp_copy)
    sample_item_2 = next(g2_tmp_copy)
    try:
        op(sample_item_1, sample_item_2)
    except TypeError:
        raise TypeError(f"Operator '{op.__name__}' cannot be applied to items produced by {g1} and {g2} "
                        f"(which have type {type(sample_item_1)} and {type(sample_item_2)}, respectively)")


def add_generators(self, other):
    check_that_operator_can_be_applied_to_produces_items(add, self, other)
    return Apply(add, self, as_tohu_generator(other))


def radd_generators(self, other):
    check_that_operator_can_be_applied_to_produces_items(add, other, self)
    return Apply(add, as_tohu_generator(other), self)


def mul_generators(self, other):
    check_that_operator_can_be_applied_to_produces_items(mul, self, other)
    return Apply(mul, self, as_tohu_generator(other))


def rmul_generators(self, other):
    check_that_operator_can_be_applied_to_produces_items(mul, other, self)
    return Apply(mul, as_tohu_generator(other), self)


def eq_generators(self, other):
    check_that_operator_can_be_applied_to_produces_items(eq, self, other)
    return Apply(eq, self, as_tohu_generator(other))


def lt_generators(self, other):
    check_that_operator_can_be_applied_to_produces_items(lt, self, other)
    return Apply(lt, self, as_tohu_generator(other))


def le_generators(self, other):
    check_that_operator_can_be_applied_to_produces_items(le, self, other)
    return Apply(le, self, as_tohu_generator(other))


def gt_generators(self, other):
    check_that_operator_can_be_applied_to_produces_items(gt, self, other)
    return Apply(gt, self, as_tohu_generator(other))


def ge_generators(self, other):
    check_that_operator_can_be_applied_to_produces_items(ge, self, other)
    return Apply(ge, self, as_tohu_generator(other))


# Patch TohuBaseGenerator with the new methods
TohuBaseGenerator.__add__ = add_generators
TohuBaseGenerator.__radd__ = radd_generators
TohuBaseGenerator.__mul__ = mul_generators
TohuBaseGenerator.__rmul__ = rmul_generators
TohuBaseGenerator.__eq__ = eq_generators
TohuBaseGenerator.__lt__ = lt_generators
TohuBaseGenerator.__le__ = le_generators
TohuBaseGenerator.__gt__ = gt_generators
TohuBaseGenerator.__ge__ = ge_generators



def split_geolocation(self):
    attributes = ['lon', 'lat'] + self.include_attributes
    return tuple(GetAttribute(self, attr_name) for attr_name in attributes)

GeoJSONGeolocation.split = split_geolocation



import textwrap
from bidict import bidict
from itertools import count

from .logging import logger
from .primitive_generators import PrimitiveGenerator
from .derived_generators import Apply, GetAttribute, Lookup, SelectOneDerived

__all__ = ['SpawnContext']


class NoExistingSpawn(Exception):
    """
    Custom exception
    """


class SpawnContext:

    def __init__(self):
        self.templates = bidict()  # mapping {name -> field_generator_template}
        self.spawns = {}  # mapping {name -> field_generator}
        self.anonymous_spawns = []  # names of anonymously spawned generators
        self.cnt_anonymous = count()

    def __repr__(self):
        return textwrap.dedent(f"""
            <SpawnContextCG:
                templates: {dict(self.templates)}
                spawns:    {dict(self.spawns)}
                anonymous: {self.anonymous_spawns}
            >""")

    @property
    def named_spawns(self):
        return {name: g for (name, g) in self.spawns.items() if name not in self.anonymous_spawns}

    def get_existing_spawn(self, g_tpl):
        try:
            existing_name = self.templates.inv[g_tpl]
            return self.spawns[existing_name]
        except KeyError:
            logger.debug(f"No existing spawn for {g_tpl}")
            raise NoExistingSpawn()

    def spawn_template(self, g_tpl, *, name):
        if name is None:
            try:
                name = self.templates.inv[g_tpl]
            except KeyError:
                name = f'ANONYMOUS_ANONYMOUS_ANONYMOUS_{next(self.cnt_anonymous)}'
                self.anonymous_spawns.append(name)
                logger.debug(f"Found anonymous field generator template: {g_tpl}")

        try:
            self.spawns[name] = self.get_existing_spawn(g_tpl)
        except NoExistingSpawn:
            if isinstance(g_tpl, PrimitiveGenerator):
                self.templates[name] = g_tpl
                self.spawns[name] = g_tpl.spawn()
            elif isinstance(g_tpl, SelectOneDerived):
                new_parent = self.spawn_template(g_tpl.parent, name=None)
                self.templates[name] = g_tpl
                self.spawns[name] = SelectOneDerived(new_parent)
            elif isinstance(g_tpl, GetAttribute):
                new_parent = self.spawn_template(g_tpl.parent, name=None)
                self.templates[name] = g_tpl
                self.spawns[name] = GetAttribute(new_parent, name=g_tpl.name)
            elif isinstance(g_tpl, Lookup):
                new_parent = self.spawn_template(g_tpl.parent, name=None)
                self.templates[name] = g_tpl
                self.spawns[name] = Lookup(new_parent, mapping=g_tpl.mapping)
            elif isinstance(g_tpl, Apply):
                new_arg_gens = []
                for gen in g_tpl.func_arg_gens_orig.arg_gens:
                    new_arg_gens.append(self.spawn_template(gen, name=None))

                new_kwarg_gens = {}
                for gen_name, gen in g_tpl.func_arg_gens_orig.kwarg_gens.items():
                    new_kwarg_gens[gen_name] = self.spawn_template(gen, name=None)

                self.templates[name] = g_tpl
                self.spawns[name] = Apply(g_tpl.func, *new_arg_gens, **new_kwarg_gens)
            else:
                raise NotImplementedError(f'g_tpl: {g_tpl}')

        # Set tohu_name for nicer debugging
        if name.startswith('ANONYMOUS_ANONYMOUS_ANONYMOUS_'):
            self.spawns[name].set_tohu_name(f'anonymous_{g_tpl.tohu_id}')
        else:
            self.spawns[name].set_tohu_name(name)

        return self.spawns[name]





import attr
import pandas as pd
import re

from .base import TohuBaseGenerator, SeedGenerator
from .logging import logger
from .spawn_context import SpawnContext

__all__ = ['CustomGenerator']


def set_item_class_name_on_custom_generator_class(cls):
    """
    Set the attribute `cls.__tohu_items_name__` to a string which defines the name
    of the namedtuple class which will be used to produce items for the custom
    generator.

    By default this will be the first part of the class name (before '...Generator'),
    for example:

        FoobarGenerator -> Foobar
        QuuxGenerator   -> Quux

    However, it can be set explicitly by the user by defining `__tohu_items_name__`
    in the class definition, for example:

        class Quux(CustomGenerator):
            __tohu_items_name__ = 'MyQuuxItem'
    """

    if '__tohu_items_name__' in cls.__dict__:
        logger.debug(f"Using item class name '{cls.__tohu_items_name__}' (derived from attribute '__tohu_items_name__')")
    else:
        m = re.match('^(.*)Generator$', cls.__name__)
        if m is not None:
            cls.__tohu_items_name__ = m.group(1)
            logger.debug(f"Using item class name '{cls.__tohu_items_name__}' (derived from custom generator name)")
        else:
            raise ValueError("Cannot derive class name for items to be produced by custom generator. "
                             "Please set '__tohu_items_name__' at the top of the custom generator's "
                             "definition or change its name so that it ends in '...Generator'")


def make_item_class(clsname, attr_names):
    """
    Parameters
    ----------
    clsname: string
        Name of the class to be created

    attr_names: list of strings
        Names of the attributes of the class to be created
    """

    item_cls = attr.make_class(clsname, {name: attr.ib() for name in attr_names}, repr=False, cmp=True)

    def new_repr(self):
        all_fields = ', '.join([f'{name}={repr(value)}' for name, value in attr.asdict(self).items()])
        return f'{clsname}({all_fields})'

    orig_eq = item_cls.__eq__

    def new_eq(self, other):
        """
        Custom __eq__() method which also allows comparisons with
        tuples and dictionaries. This is mostly for convenience
        during testing.
        """

        if isinstance(other, self.__class__):
            return orig_eq(self, other)
        else:
            if isinstance(other, tuple):
                return attr.astuple(self) == other
            elif isinstance(other, dict):
                return attr.asdict(self) == other
            else:
                return NotImplemented

    item_cls.__repr__ = new_repr
    item_cls.__eq__ = new_eq
    item_cls.keys = lambda self: attr_names
    item_cls.__getitem__ = lambda self, key: getattr(self, key)
    item_cls.as_dict = lambda self: attr.asdict(self)
    item_cls.to_series = lambda self: pd.Series(attr.asdict(self))

    return item_cls


class CustomGenerator(TohuBaseGenerator):

    def __init__(self, *args, **kwargs):
        super().__init__()

        self.orig_args = args
        self.orig_kwargs = kwargs

        self.seed_generator = SeedGenerator()
        self.field_gen_templates = self._find_field_generator_templates()

        spawn_context = SpawnContext()
        for name, g in self.field_gen_templates.items():
            spawn_context.spawn_template(g, name=name)

        self.field_gens_named = spawn_context.named_spawns
        self.field_gens_all = spawn_context.spawns

        self.__dict__.update(self.field_gens_named)

        set_item_class_name_on_custom_generator_class(self.__class__)
        self._set_field_names_and_field_gens()
        self._set_item_class()

    def _set_field_names_and_field_gens(self):
        field_gen_names = list(self.field_gens_named.keys())

        if hasattr(self, '__fields__'):
            self.field_names = self.__fields__
        else:
            self.field_names = field_gen_names

        if hasattr(self, '__fields__'):
            # sanity check
            for field_name in self.field_names:
                if field_name not in field_gen_names:
                    raise ValueError(f"Attribute __fields__ contains name which is not a named field generator: '{field_name}'")

        self.field_gens = {name: gen for (name, gen) in self.field_gens_named.items() if name in self.field_names}

    def _set_item_class(self):
        """
        cls:
            The custom generator class for which to create an item-class
        """
        clsname = self.__tohu_items_name__
        self.item_cls = make_item_class(clsname, self.field_names)

    def __next__(self):
        field_values = [next(g) for g in self.field_gens.values()]
        return self.item_cls(*field_values)

    def _find_field_generator_templates(self):
        """
        Return a dictionary of the form {name: field_generator} containing
        all tohu generators defined in the class and instance namespace
        of this custom generator.
        """
        field_gen_templates = {}

        # Extract field generators from class dict
        for name, g in self.__class__.__dict__.items():
            if isinstance(g, TohuBaseGenerator):
                field_gen_templates[name] = g.set_tohu_name(f'{name} (TPL)')

        # Extract field generators from instance dict
        for name, g in self.__dict__.items():
            if isinstance(g, TohuBaseGenerator):
                field_gen_templates[name] = g.set_tohu_name(f'{name} (TPL)')

        return field_gen_templates

    def reset(self, seed):
        super().reset(seed)
        self.seed_generator.reset(seed)
        for name, gen in self.field_gens_all.items(): # FIXME: reset all field gens
            next_seed = next(self.seed_generator)
            gen.reset(next_seed)
        return self

    def spawn(self):
        new_obj = self.__class__(*self.orig_args, **self.orig_kwargs)
        new_obj._set_random_state_from(self)

        # Explicitly set item_cls. This is necessary because due to
        # the way in which `attr` works, explicit comparisons between
        # generated items will return False even though they contain
        # the same elements (because the underlying attr classes are
        # different, so attr plays it safe).
        new_obj.item_cls = self.item_cls

        return new_obj

    def _set_random_state_from(self, other):
        self.seed_generator._set_random_state_from(other.seed_generator)

        # FIXME: should also set random state for unnamed field generators
        #        (these can occur in chains of derived generators)
        for name in self.field_gens_all.keys():
            self.field_gens_all[name]._set_random_state_from(other.field_gens_all[name])



import hashlib

from abc import ABCMeta, abstractmethod
from itertools import islice
from random import Random
from tqdm import tqdm

from .item_list import ItemList
from .logging import logger

__all__ = ['SeedGenerator', 'TohuBaseGenerator']


class TohuBaseGenerator(metaclass=ABCMeta):
    """
    Base class for all of tohu's generators.
    """

    def __init__(self):
        self._clones = []
        self.tohu_name = None

    def __repr__(self):
        clsname = self.__class__.__name__
        name = '' if self.tohu_name is None else f'{self.tohu_name}: '
        return f'<{name}{clsname} (id={self.tohu_id})>'

    def __format__(self, fmt):
        return self.__repr__()

    def set_tohu_name(self, tohu_name):
        """
        Set this generator's `tohu_name` attribute.

        This is mainly useful for debugging where one can temporarily
        use this at the end of generator definitions to set a name
        that will be displayed in debugging messages. For example:

            g1 = SomeGeneratorClass().set_tohu_name('g1')
            g2 = SomeGeneratorClass().set_tohu_name('g2')
        """
        self.tohu_name = tohu_name
        return self

    @property
    def tohu_id(self):
        """
        Return (truncated) md5 hash representing this generator.
        We truncate the hash simply for readability, as this is
        purely intended for debugging purposes and the risk of
        any collisions will be negligible.
        """
        myhash = hashlib.md5(str(id(self)).encode()).hexdigest()
        return myhash[:12]

    def __iter__(self):
        return self

    def register_clone(self, clone):
        self._clones.append(clone)

        # If possible, set the clone's tohu_name for easier debugging
        if self.tohu_name is not None:
            clone.set_tohu_name(f"{self.tohu_name} (clone #{len(self._clones)})")

        if len(self._clones) != len(set(self._clones)):
            raise RuntimeError(f"Duplicate clone added: {self}  -->  {clone}")

    @abstractmethod
    def __next__(self):
        raise NotImplementedError("Class {} does not implement method '__next__'.".format(self.__class__.__name__))

    @abstractmethod
    def reset(self, seed):
        logger.debug(f'Resetting {self} (seed={seed})')

        # Print debugging messages in bulk before actually
        # resetting clones because it makes them easier to
        # read in the debugging output.
        if self._clones != []:
            logger.debug('  Will also reset the following clones:')
            for c in self._clones:
                logger.debug(f'   - {c}')

        for c in self._clones:
            c.reset(seed)

    @abstractmethod
    def _set_random_state_from(self, other):
        """
        This helper method sets the internal random state
        of `self` to the same state that `other` is in. This
        ensures that afterwards any the two generators `self`
        and `other` produce the same elements in the same
        order (even though otherwise they remain independent).
        This is used internally when spawning generators.
        """
        raise NotImplementedError("Class {} does not implement method 'spawn'.".format(self.__class__.__name__))

    @abstractmethod
    def spawn(self):
        raise NotImplementedError("Class {} does not implement method 'spawn'.".format(self.__class__.__name__))

    def clone(self):
        c = self.spawn()
        self.register_clone(c)
        return c

    def generate(self, num, *, seed=None, progressbar=False):
        """
        Return sequence of `num` elements.

        If `seed` is not None, the generator is reset
        using this seed before generating the elements.
        """
        if seed is not None:
            self.reset(seed)

        items = islice(self, num)
        if progressbar:
            items = tqdm(items, total=num)

        item_list = [x for x in items]

        #logger.warning("TODO: initialise ItemList with random seed!")
        return ItemList(item_list, num)


class SeedGenerator:
    """
    This class is used in custom generators to create a collection of
    seeds when reset() is called, so that each of the constituent
    generators can be re-initialised with a different seed in a
    reproducible way.

    Note: This is almost identical to the `tohu.Integer` generator, but we
    need a version which does *not* inherit from `TohuUltraBaseGenerator`,
    otherwise the automatic item class creation in `CustomGeneratorMeta`
    gets confused.
    """

    def __init__(self):
        self.randgen = Random()
        self.minval = 0
        self.maxval = 2**32 - 1

    def reset(self, seed):
        self.randgen.seed(seed)

    def __iter__(self):
        return self

    def __next__(self):
        return self.randgen.randint(self.minval, self.maxval)

    def _set_random_state_from(self, other):
        self.randgen.setstate(other.randgen.getstate())


import io
import logging
import numpy as np
import os
import re
import pandas as pd
from operator import attrgetter
from sqlalchemy import create_engine, inspect
from sqlalchemy.schema import CreateSchema

from .utils import explode_columns

logger = logging.getLogger('tohu')


def _generate_csv_header_line(*, header_names, header_prefix='', header=True, sep=',', newline='\n'):
    """
    Helper function to generate a CSV header line depending on
    the combination of arguments provided.
    """
    if isinstance(header, str):  # user-provided header line
        header_line = header + newline
    else:
        if not (header is None or isinstance(header, bool)):
            raise ValueError(f"Invalid value for argument `header`: {header}")
        else:
            if header:
                header_line = header_prefix + sep.join(header_names) + newline
            else:
                header_line = ""
    return header_line


def _extract_schema_if_given(table_name):
    """
    Return a pair (schema, table) derived from the given `table_name`
    (anything before the first '.' if the name contains one; otherwise
    the return value of `schema` is None).

    Examples:

        >>> _extract_schema_if_given('some_schema.my_table')
        ('some_schema', 'my_table')

        >>> _extract_schema_if_given('my_awesome_table')
        (None, 'my_awesome_table')
    """
    pattern = '^(([^.]+)\.)?(.+)$'
    m = re.match(pattern, table_name)
    schema, table_name = m.group(2), m.group(3)
    return schema, table_name


class ItemList:

    def __init__(self, items, num):
        self.items = items if isinstance(items, list) else list(items)
        self.num = num
        self.randstate = np.random.RandomState()

    def __repr__(self):
        return f"<ItemList containing {self.num} items>"

    def __eq__(self, other):
        return self.items == other

    def __len__(self):
        return self.num

    def __getitem__(self, idx):
        return self.items[idx]

    def __iter__(self):
        return iter(self.items)

    def reset(self, seed):
        """
        Reset the internal random state of this ItemList.
        This is useful to generate reproducible subsamples.
        """
        if seed is None:
            return
        self.randstate.seed(seed)

    def subsample(self, *, num=None, p=None, seed=None):
        self.reset(seed)

        if num is None and p is None:
            raise ValueError("Exactly one of the arguments `num`, `p` must be given.")

        if num is not None:
            if num > self.num:
                raise ValueError(f"Subsample cannot be larger than the original sample of size {self.num}")
            return ItemList(self.randstate.choice(self.items, size=num, replace=False), num=num)
        elif p is not None:
            if p < 0 or p > 1.0:
                raise ValueError(f"The value of p must be in the range [0, 1]. Got: p={p}")
            subsample = np.array(self.items)[self.randstate.random_sample(self.num) < p]
            return subsample
        else:
            raise ValueError("Arguments `num` and `p` are mutually exclusive - only one of them may be specified.")

    def to_df(self, fields=None, fields_to_explode=None):
        """
        Export items as rows in a pandas dataframe table.

        Parameters
        ----------

        fields: list or dict
            List of field names to export, or dictionary mapping output column names
            to attribute names of the generators.

            Examples:
               fields=['field_name_1', 'field_name_2']
               fields={'COL1': 'field_name_1', 'COL2': 'field_name_2'}

        fields_to_explode: list or None
            Optional list of field names where each entry (which must itself be a sequence)
            is to be "exploded" into separate rows.

        """
        if isinstance(fields, (list, tuple)):
            fields = {name: name for name in fields}

        assert fields_to_explode is None or isinstance(fields_to_explode, (list, tuple))
        if fields_to_explode is None:
            fields_to_explode = []

        if fields is None:
            colnames_to_export = list(self.items[0].as_dict().keys())  # hack! the field names should perhaps be passed in during initialisation?
        else:
            colnames_to_export = list(fields.keys())

        if not set(fields_to_explode).issubset(colnames_to_export):
            raise ValueError(
                "All fields to explode must occur as column names. "
                f"Got field names: {fields_to_explode}. Column names: {list(fields.keys())}"
            )

        if fields is None:
            # New version (much faster!, but needs cleaning up)
            import attr
            df = pd.DataFrame([attr.astuple(x) for x in self.items], columns=colnames_to_export)
            # Old version:
            #return pd.DataFrame([x.to_series() for x in self.items])
        else:
            # New version (much faster!)

            def make_attrgetter(attr_name_new, attr_name, fields_to_explode):
                # TODO: this needs cleaning up!
                if attr_name_new in fields_to_explode and '.' in attr_name:
                    attr_name_first_part, attr_name_rest = attr_name.split('.', maxsplit=1)

                    def func(row):
                        foo_items = attrgetter(attr_name_first_part)(row)
                        return [attrgetter(attr_name_rest)(x) for x in foo_items]

                    return func
                else:
                    return attrgetter(attr_name)

            attr_getters = [make_attrgetter(attr_name_new, attr_name, fields_to_explode) for attr_name_new, attr_name in fields.items()]
            try:
                df = pd.DataFrame([tuple(func(x) for func in attr_getters) for x in self.items], columns=colnames_to_export)
            except AttributeError as exc:
                msg = (
                    "Could not export to dataframe. Did you forget to pass any fields "
                    "which contain sequences within the 'fields_to_explode' argument?. "
                    f"The original error message was: \"{exc}\""
                )
                raise AttributeError(msg)

        if fields_to_explode != []:
            # TODO: add sanity checks to avoid unwanted behaviour (e.g. that all columns
            # to be exploded must have the same number of elements in each entry?)
            df = explode_columns(df, fields_to_explode)

        return df

    def to_csv(self, filename=None, *, fields=None, fields_to_explode=None, append=False, header=True, header_prefix='', sep=',', newline='\n'):
        """
        Parameters
        ----------
        filename: str or None
            The file to which output will be written. By default, any existing content is
            overwritten. Use `append=True` to open the file in append mode instead.
            If filename is None, the generated CSV output is returned instead of written
            to a file.
        fields: list or dict
            List of field names to export, or dictionary mapping output column names
            to attribute names of the generators.

            Examples:
               fields=['field_name_1', 'field_name_2']
               fields={'COL1': 'field_name_1', 'COL2': 'field_name_2'}
        fields_to_explode: list
            Optional list of field names where each entry (which must itself be a sequence)
            is to be "exploded" into separate rows. (*Note:* this is not supported yet for CSV export.)
        append: bool
            If `True`, open the file in 'append' mode to avoid overwriting existing content.
            Default is `False`, i.e. any existing content will be overwritten.
            This argument only has an effect if `filename` is given (i.e. if output happens
            to a file instead of returning a CSV string).
        header: bool or str or None
            If `header=False` or `header=None` then no header line will be written.
            If `header` is a string then this string will be used as the header line.
            If `header=True` then a header line will be automatically generated from
            the field names of the custom generator.
        header_prefix: str
            If `header=True` then the auto-generated header line will be prefixed
            with `header_prefix` (otherwise this argument has no effect). For example,
            set `header_prefix='#'` to make the header line start with '#'. Default: ''
        sep: str
            Field separator to use in the output. Default: ','
        newline: str
            Line terminator to use in the output. Default: '\n'

        Returns
        -------
        The return value depends on the value of `filename`.
        If `filename` is given, writes the output to the file and returns `None`.
        If `filename` is `None`, returns a string containing the CSV output.
        """
        assert isinstance(append, bool)

        if fields is None:
            raise NotImplementedError("TODO: derive field names automatically from the generator which produced this item list")

        if fields_to_explode is not None:
            raise NotImplementedError("TODO: the 'fields_to_explode' argument is not supported for CSV export yet.")

        if isinstance(fields, (list, tuple)):
            fields = {name: name for name in fields}

        header_line = _generate_csv_header_line(header=header, header_prefix=header_prefix, header_names=fields.keys(), sep=sep, newline=newline)

        if filename is not None:
            # ensure parent directory of output file exits
            dirname = os.path.dirname(os.path.abspath(filename))
            if not os.path.exists(dirname):
                os.makedirs(dirname)
        file_or_string = open(filename, 'a' if append else 'w') if (filename is not None) else io.StringIO()

        retval = None
        attr_getters = [attrgetter(attr_name) for attr_name in fields.values()]
        try:

            file_or_string.write(header_line)

            for x in self.items:
                line = sep.join([format(func(x)) for func in attr_getters]) + newline
                file_or_string.write(line)

            if filename is None:
                retval = file_or_string.getvalue()

        finally:
            file_or_string.close()

        return retval

    def to_sql(self, url, table_name, *, schema=None, fields=None, fields_to_explode=None, if_exists="fail", dtype=None):
        """
        Export items as rows in a PostgreSQL table.

        Parameters
        ----------

        url: string
            Connection string to connect to the database.
            Example: "postgresql://postgres@127.0.0.1:5432/testdb"

        table_name: string
            Name of the database table. Note that if this name contains a dot ('.')
            and `schema` is not specified, the first part of the name before the dot
            will be interpreted as the schema name.

        schema : string, optional
            Specify the schema (if database flavor supports this). If None,
            use default schema or derive the schema name from `table_name`.

        fields: list or dict
            List of field names to export, or dictionary mapping output column names
            to attribute names of the generators.

            Examples:
               fields=['field_name_1', 'field_name_2']
               fields={'COL1': 'field_name_1', 'COL2': 'field_name_2'}

        fields_to_explode: list or None
            Optional list of field names where each entry (which must itself be a sequence)
            is to be "exploded" into separate rows.

        if_exists : {'fail', 'do_nothing', 'replace', 'append'}, default 'fail'
            - fail: If table exists, raise an error.
            - do_nothing: If table exists, do nothing and immediately return.
            - replace: If table exists, drop it, recreate it, and insert data.
            - append: If table exists, insert data. Create if does not exist.

        dtype : dict, optional
            Specifying the datatype for columns. The keys should be the column
            names and the values should be the SQLAlchemy types or strings for
            the sqlite3 legacy mode. This is passed through to pandas.DataFrame.to_sql().
        """
        if schema is None:
            schema, table_name = _extract_schema_if_given(table_name)

        engine = create_engine(url)
        ins = inspect(engine)

        if schema is not None and schema not in ins.get_schema_names():
            logger.debug(f"Creating non-existing schema: '{schema}'")
            engine.execute(CreateSchema(schema))

        if table_name in ins.get_table_names(schema=schema) and if_exists == 'do_nothing':
            logger.debug("Table already exists (use if_exists='replace' or if_exists='append' to modify it).")
            return

        if if_exists == 'do_nothing':
            # we handled the 'do nothing' case above; change to an option that pandas will understand
            if_exists = 'fail'

        with engine.begin() as conn:
            self.to_df(fields=fields, fields_to_explode=fields_to_explode).to_sql(
                table_name, conn, schema=schema, index=False, if_exists=if_exists, dtype=dtype)



import hashlib

from abc import ABCMeta, abstractmethod
from itertools import islice
from random import Random
from tqdm import tqdm

from .item_list import ItemList
from .logging import logger

__all__ = ['SeedGenerator', 'TohuBaseGenerator', 'PrimitiveGenerator']


class SeedGenerator:
    """
    This class is used in custom generators to create a collection of
    seeds when reset() is called, so that each of the constituent
    generators can be re-initialised with a different seed in a
    reproducible way.

    Note: This is almost identical to the `tohu.Integer` generator, but we
    need a version which does *not* inherit from `TohuBaseGenerator`,
    otherwise the automatic item class creation in custom generators
    gets confused.
    """

    def __init__(self):
        self.randgen = Random()
        self.minval = 0
        self.maxval = 2**32 - 1

    def reset(self, seed):
        self.randgen.seed(seed)
        return self

    def __iter__(self):
        return self

    def __next__(self):
        return self.randgen.randint(self.minval, self.maxval)

    def _set_random_state_from(self, other):
        self.randgen.setstate(other.randgen.getstate())


class TohuCloneError(Exception):
    """
    Custom exception
    """


class TohuBaseGenerator(metaclass=ABCMeta):
    """
    Base class for all of tohu's generators.
    """

    def __init__(self, *args, **kwargs):
        self.tohu_name = None
        self.owner = None
        self.parent = None
        self.clones = []
        self.input_generators = []
        self.seed_generator = SeedGenerator()
        self.is_custom_generator_template = False
        self._max_value = None

    def __repr__(self):
        clsname = self.__class__.__name__
        name = '' if self.tohu_name is None else (f'{self.tohu_name}: ' if not self.is_custom_generator_template else f'{self.tohu_name} (TPL): ')
        owned_by = '' if self.owner is None else f' [owned by: {self.owner}] '
        return f'<{name}{clsname} (id={self.tohu_id}){owned_by}>'

    def __format__(self, fmt):
        return self.__repr__()

    def set_tohu_name(self, tohu_name):
        """
        Set this generator's `tohu_name` attribute.

        This is mainly useful for debugging where one can temporarily
        use this at the end of generator definitions to set a name
        that will be displayed in debugging messages. For example:

            g1 = SomeGeneratorClass().set_tohu_name('g1')
            g2 = SomeGeneratorClass().set_tohu_name('g2')
        """
        self.tohu_name = tohu_name
        return self

    @property
    def tohu_id(self):
        """
        Return (truncated) md5 hash representing this generator.
        We truncate the hash simply for readability, as this is
        purely intended for debugging purposes and the risk of
        any collisions will be negligible.
        """
        myhash = hashlib.md5(str(id(self)).encode()).hexdigest()
        return myhash[:12]

    def __iter__(self):
        return self

    @abstractmethod
    def __next__(self):
        raise NotImplementedError(f"Class {self.__class__.__name__} does not implement method '__next__'.")

    @property
    def max_value(self):
        return self._max_value

    @max_value.setter
    def max_value(self, value):
        if self._max_value is None:
            self._max_value = value
        else:
            if value is not None:
                raise ValueError(f"Trying to set attribute max_value={value} but it already has value {self._max_value}")

    @abstractmethod
    def reset(self, seed):
        """
        Reset this generator's seed generator and any clones.
        """
        logger.debug(f'Resetting {self} (seed={seed})')
        self.seed_generator.reset(seed)

        for c in self.clones:
            c.reset(seed)

    def generate(self, num, *, seed=None, progressbar=False):
        """
        Return sequence of `num` elements.

        If `seed` is not None, the generator is reset
        using this seed before generating the elements.
        """
        if seed is not None:
            self.reset(seed)

        items = islice(self, num)
        if progressbar:
            items = tqdm(items, total=num)

        item_list = [x for x in items]

        #logger.warning("TODO: initialise ItemList with random seed!")
        return ItemList(item_list, num)

    @abstractmethod
    def _set_random_state_from(self, other):
        logger.debug(f"Setting internal state of {self} (from {other})")
        self.seed_generator._set_random_state_from(other.seed_generator)

    @abstractmethod
    def spawn(self, spawn_mapping=None):
        raise NotImplementedError("Class {} does not implement method 'spawn'.".format(self.__class__.__name__))

    def clone(self):
        """
        Return an exact copy of this generator which behaves the same way
        (i.e., produces the same elements in the same order) and which is
        automatically reset whenever the original generator is reset.
        """
        c = self.spawn()
        self.register_clone(c)
        c.register_parent(self)
        return c

    def register_clone(self, clone):
        self.clones.append(clone)

        # If possible, set the clone's tohu_name for easier debugging
        if self.tohu_name is not None:
            clone.set_tohu_name(f"{self.tohu_name} (clone #{len(self.clones)})")

        if len(self.clones) != len(set(self.clones)):
            raise RuntimeError(f"Duplicate clone added: {self}  -->  {clone}")

    def unregister_clone(self, clone):
        try:
            self.clones.remove(clone)
        except ValueError:
            raise TohuCloneError(f"Cannot unregister clone {clone} because it is not a clone of parent {self}")

    def register_parent(self, parent):
        self.parent = parent


class PrimitiveGenerator(TohuBaseGenerator):
    """
    Base class for all primitive generators
    """


"""
This module is not meant to be imported directly.
Its purpose is to patch the TohuBaseGenerator class
so that its special methods __add__, __mul__ etc.
support other generators as arguments.
"""

from operator import add, mul, gt, ge, lt, le, eq

from .base import TohuBaseGenerator
from .primitive_generators import GeoJSONGeolocation, SelectOnePrimitive, Timestamp, as_tohu_generator
from .derived_generators import Apply, GetAttribute, SelectOneDerived
from .utils import identity

__all__ = []


def add_generators(self, other):
    return Apply(add, self, as_tohu_generator(other))


def radd_generators(self, other):
    return Apply(add, as_tohu_generator(other), self)


def mul_generators(self, other):
    return Apply(mul, self, as_tohu_generator(other))


def rmul_generators(self, other):
    return Apply(mul, as_tohu_generator(other), self)


def eq_generators(self, other):
    return Apply(eq, self, as_tohu_generator(other))


def lt_generators(self, other):
    return Apply(lt, self, as_tohu_generator(other))


def le_generators(self, other):
    return Apply(le, self, as_tohu_generator(other))


def gt_generators(self, other):
    return Apply(gt, self, as_tohu_generator(other))


def ge_generators(self, other):
    return Apply(ge, self, as_tohu_generator(other))


# Patch TohuBaseGenerator with the new methods
TohuBaseGenerator.__add__ = add_generators
TohuBaseGenerator.__radd__ = radd_generators
TohuBaseGenerator.__mul__ = mul_generators
TohuBaseGenerator.__rmul__ = rmul_generators
TohuBaseGenerator.__eq__ = eq_generators
TohuBaseGenerator.__lt__ = lt_generators
TohuBaseGenerator.__le__ = le_generators
TohuBaseGenerator.__gt__ = gt_generators
TohuBaseGenerator.__ge__ = ge_generators


def getattribute_generators(self, name):

    if name == '__isabstractmethod__':
        # Special case which is needed because TohuUltraBaseMeta is
        # derived from ABCMeta and it uses '__isabstractmethod__'
        # to check for abstract methods.
        #
        # TODO: This check should probably be moved to TohuUltraBaseGenerator somewhere.
        return

    if name == '_ipython_canary_method_should_not_exist_':
        # Special case which is needed because IPython uses this attribute internally.
        raise NotImplementedError("Special case needed for IPython")

    return GetAttribute(self, name)

SelectOnePrimitive.__getattr__ = getattribute_generators
SelectOneDerived.__getattr__ = getattribute_generators


def split_geolocation(self):
    attributes = ['lon', 'lat'] + self.include_attributes
    return tuple(GetAttribute(self, attr_name) for attr_name in attributes)

GeoJSONGeolocation.split = split_geolocation
